// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "aggregation.hh"

#include "packet-dns.hh"
#include "util.hh"
#include "value.hh"

#include <ctime>
#include <iostream>
#include <queue>
#include <sstream>

namespace dns_analyzer {

    using std::cout;
    using std::cerr;
    using std::deque;
    using std::endl;
    using std::ios;
    using std::istringstream;
    using std::pair;
    using std::priority_queue;
    using std::time_t;

    typedef pair<variable_type::group_values, combined_stat_type> top_queue_value_type;

    struct greater_pair_second_type {
        bool operator()(top_queue_value_type const &lhs,
                        top_queue_value_type const &rhs) const
        {
            return lhs.second.sum() > rhs.second.sum();
        }
    };

    static void
    print_group_by(ostream &out, variable_type const &data) {
        variable_type::group_names const &names = data.group_by();
        for (size_t i = 0; i < names.size(); ++i) {
            out << names.at(i) << ' ';
        }
        out << "count\n";
    }

    static void
    print_stat(ostream &out,
               variable_type::group_stat_type::value_type const &value)
    {
        for (size_t j = 0; j < value.first.size(); ++j) {
            ostringstream temp;
            temp << value.first.at(j);
            write_quoted_string(out, temp.str());
            if (j < value.first.size() - 1) {
                out << ' ';
            }
        }
        out << ' ';
        ios::fmtflags old = out.setf(ios::fixed, ios::floatfield);
        out.precision(0);
        out << value.second.sum();
        out.setf(old);
        out << '\n';
    }

    static void
    print_top_n(ostream &out, size_t n, variable_type const &data) {
        stat_type global(data.global_stats());
        out.setf(ios::dec, ios::basefield);
        out << "Top " << n << ": " << global << endl;

        print_group_by(out, data);

        typedef priority_queue<top_queue_value_type, deque<top_queue_value_type>, greater_pair_second_type> queue_type;

        queue_type top_queue;
        variable_type::group_stat_type const &stats = data.stats();
        for (variable_type::group_stat_type::const_iterator it = stats.begin();
             it != stats.end();
             ++it)
        {
            variable_type::group_stat_type::value_type const &value = *it;
            if (top_queue.size() < n) {
                top_queue.push(value);
            } else {
                top_queue_value_type const &lowest = top_queue.top();
                if (value.second.sum() > lowest.second.sum()) {
                    top_queue.pop();
                    top_queue.push(value);
                }
            }
        }

        while (!top_queue.empty()) {
            print_stat(out, top_queue.top());
            top_queue.pop();
        }
    }

    static void
    print_all_data(ostream &out, variable_type const &data) {
        print_group_by(out, data);

        variable_type::group_stat_type const &stats = data.stats();
        for (variable_type::group_stat_type::const_iterator it = stats.begin();
             it != stats.end();
             ++it)
        {
            print_stat(out, *it);
        }
    }

    aggregation_listener_type::aggregation_listener_type()
        : _query_count(0),
          _response_count(0),
          _prune_count(0),
          _top_count(10),
          _print_all(false),
          _last_update_timestamp(0.0),
          _stat_config(),
          _fields(),
          _data()
    {
    }

    aggregation_listener_type::~aggregation_listener_type() {
    }

    void
    aggregation_listener_type::initialize(analyzer_type &/*analyzer*/) {
        variable_type::group_names column_names;
        for (vector<packet_field_type>::const_iterator it = _fields.begin();
             it != _fields.end();
             ++it)
        {
            packet_field_type const &column = *it;
            column_names.push_back(column.to_string());
        }
        _data.group_by(column_names);
    }

    void
    aggregation_listener_type::analyze(analyzer_type &analyzer,
                                       timeval const &timestamp,
                                       network_packet const &packet)
    {
        if (dns_packet const *dns = find_payload<dns_packet>(&packet)) {
            if (dns->is_query()) {
                handle_dns_query(analyzer, timestamp, packet);
            } else {
                handle_dns_response(analyzer, timestamp, packet);
            }
        }
    }

    void
    aggregation_listener_type::finalize(analyzer_type &/*analyzer*/) {
        print_results(cout);
    }

    void
    aggregation_listener_type::print_results(ostream &out) {
        out.setf(ios::dec, ios::basefield);
        out << "Query count: " << _query_count << "\n"
            << "Response count: " << _response_count << "\n";

        if (_top_count > 0) {
            print_top_n(out, _top_count, _data);
        }
        if (_print_all) {
            print_all_data(out, _data);
        }
    }

    void
    aggregation_listener_type::set_fields(packet_field_list_type const &fields) {
        _fields = fields;
    }

    void
    aggregation_listener_type::set_interval_times(
        combined_stat_config_type const *config)
    {
        if (config == 0) {
            _stat_config.reset(0);
        } else {
            _stat_config.reset(new combined_stat_config_type(*config));
        }
        _data.configure(_stat_config);
    }

    size_t
    aggregation_listener_type::packet_count() const {
        return _query_count + _response_count;
    }

    void
    aggregation_listener_type::handle_packet(analyzer_type &analyzer,
                                             timeval const &timestamp,
                                             network_packet const &packet)
    {
        float ts = (float) timeval_to_double(timestamp);

        variable_type::group_values values;
        values.reserve(_fields.size());
        for (vector<packet_field_type>::const_iterator it = _fields.begin();
             it != _fields.end();
             ++it)
        {
            packet_field_type const &column = *it;
            values.push_back(column.get(packet));
        }
        _data.add(values, 1.0);

        dns_packet const *dns = find_payload<dns_packet>(&packet);

        if (dns->question_count() == 1) {
            resource_record_type question(*dns->begin());
            if (question.type() > ns_t_any) {
                analyzer.insert_interesting(packet);
            }
        } else {
            analyzer.insert_interesting(packet);
        }

        if (_stat_config
            && ((ts - _last_update_timestamp)
                >= _stat_config->intervals().at(0)))
        {
            update_variables(ts);
        }

        if (_prune_count != 0 && (packet_count() % _prune_count) == 0) {
            prune_variables();
        }
    }

    void
    aggregation_listener_type::handle_dns_response(analyzer_type &analyzer,
                                                   timeval const &timestamp,
                                                   network_packet const &packet)
    {
        if (handle_responses()) {
            ++_response_count;
            handle_packet(analyzer, timestamp, packet);
        }
    }

    void
    aggregation_listener_type::handle_dns_query(analyzer_type &analyzer,
                                                timeval const &timestamp,
                                                network_packet const &packet)
    {
        if (handle_queries()) {
            ++_query_count;
            handle_packet(analyzer, timestamp, packet);
        }
    }

    void
    aggregation_listener_type::prune_variables() {
        _data.prune(_prune_count);
    }

    void
    aggregation_listener_type::update_variables(float timestamp) {
        _data.update(timestamp);
        _last_update_timestamp = timestamp;
    }
}
