// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef AGGREGATION_HH
#define AGGREGATION_HH

#include "analyzer.hh"

#include "pointers.hh"
#include "variable.hh"

#include <iosfwd>
#include <vector>

namespace dns_analyzer {

    using std::ostream;
    using std::vector;

    class value_ref;
    class packet_field_type;
    typedef vector<packet_field_type> packet_field_list_type;

    /**
     * Aggregate packet data based on a run-time configurable set of
     * fields.
     */
    class aggregation_listener_type : public listener_type {
    public:
        aggregation_listener_type();
        virtual ~aggregation_listener_type();

        virtual void initialize(analyzer_type &analyzer);
        virtual void analyze(analyzer_type &analyzer,
                             timeval const &timestamp,
                             network_packet const &packet);
        virtual void finalize(analyzer_type &analyzer);

        /**
         * Set the fields to aggregate on.
         *
         * @param fields the list of fields.
         */
        void set_fields(packet_field_list_type const &fields);

        /**
         * Configure the intervals used to gather throughput
         * information based on the field aggregation.
         *
         * @param config the list of time intervals.
         */
        void set_interval_times(combined_stat_config_type const *config);

        /**
         * Set the number of packets to analyze before pruning the
         * aggregation table to reduce memory usage.
         *
         * @param value the number of packets to analyze before
         * pruning.
         */
        void prune_count(size_t value) { _prune_count = value; }

        /**
         * The number of packets to analyze before pruning.
         */
        size_t prune_count() const { return _prune_count; }

        /**
         * Set the number of top combinations to print after analyzing
         * all the packets.
         *
         * @param value the number of top combinations to print.
         */
        void top_count(size_t value) { _top_count = value; }

        /**
         * The number of top combinations to print.
         */
        size_t top_count() const { return _top_count; }

        /**
         * Enable or disable printing all field combinations with
         * aggregation count after completing the analysis.
         *
         * @param value enable flag.
         */
        void print_all(bool value) { _print_all = value; }

        /**
         * Is print all field combinations with aggregation count data
         * enabled?
         */
        bool print_all() const { return _print_all; }

        /**
         * Enable or disable analyzing of DNS responses.
         *
         * @param value enable flag.
         */
        void handle_responses(bool value) { _handle_responses = value; }

        /**
         * Is analyzing of DNS responses enabled?
         */
        bool handle_responses() const { return _handle_responses; }

        /**
         * Enable or disable analyzing of DNS queries.
         *
         * @param value enable flag.
         */
        void handle_queries(bool value) { _handle_queries = value; }

        /**
         * Is analyzing of DNS queries enabled?
         */
        bool handle_queries() const { return _handle_queries; }
    private:
        size_t _query_count;
        size_t _response_count;
        size_t _prune_count;
        size_t _top_count;
        size_t _print_all;
        bool _handle_responses;
        bool _handle_queries;
        float _last_update_timestamp;
        scoped_ptr<combined_stat_config_type const> _stat_config;
        vector<packet_field_type> _fields;
        variable_type _data;

        void print_results(ostream &out);

        size_t packet_count() const;

        void handle_packet(analyzer_type &analyzer,
                           timeval const &timestamp,
                           network_packet const &packet);
        void handle_dns_response(analyzer_type &analyzer,
                                 timeval const &timestamp,
                                 network_packet const &packet);
        void handle_dns_query(analyzer_type &analyzer,
                              timeval const &timestamp,
                              network_packet const &packet);

        void prune_variables();
        void update_variables(float timestamp);
    };
}

#endif // AGGREGATION_HH
