// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "analyzer.hh"

#include "packet.hh"
#include "pcap.hh"
#include "util.hh"

#include <iostream>
#include <sstream>
#include <stdexcept>

namespace dns_analyzer {

    using std::cerr;
    using std::endl;
    using std::exception;
    using std::out_of_range;
    using std::runtime_error;

    float current_packet_timestamp = 0.0;

    listener_type::~listener_type() {}

    static timeval
    operator-(timeval const &lhs, timeval const &rhs) {
        timeval result;
        result.tv_sec = lhs.tv_sec - rhs.tv_sec;
        result.tv_usec = lhs.tv_usec - rhs.tv_usec;
        if (result.tv_usec < 0) {
            // borrow
            --result.tv_sec;
            result.tv_usec += 1000000;
        }
        return result;
    }

    static bool
    operator<(timeval const &lhs, timeval const &rhs) {
        return (lhs.tv_sec < rhs.tv_sec
                || (lhs.tv_sec == rhs.tv_sec
                    && lhs.tv_usec < rhs.tv_usec));
    }

    analyzer_type::analyzer_type(string const &program_name)
        : _packets(),
          _dumper(_packets),
          _program_name(program_name),
          _anonymize(false),
          _interesting(false),
          _dump_interesting(""),
          _trace_filenames(),
          _listeners(),
          _packet_count(0)
    {
    }

    void
    analyzer_type::display_exception(char const *msg, exception const &ex, network_packet const &packet) {
        _dumper.dump(const_cast<pcap_pkthdr *>(&packet.data().header()),
                     packet.data().begin());
        cerr << "EXCEPTION: " << msg << ": " << ex.what() << ": ";
        try {
            packet.print_info(cerr);
        } catch (...) {
            cerr << "<EXCEPTION WHILE PARSING PACKET>";
        }
        cerr << endl;
    }

    void
    analyzer_type::insert_trace_filename(string const &filename) {
        _trace_filenames.insert(filename);
    }

    void
    analyzer_type::insert_listener(listener_type *listener) {
        _listeners.insert(listener);
    }

    void
    analyzer_type::erase_listener(listener_type *listener) {
        _listeners.erase(listener);
    }

    void
    analyzer_type::insert_interesting(network_packet const &packet) {
        if (interesting()) {
            _interesting_packets.push_back(packet.deep_copy());
            _dumper.dump(const_cast<pcap_pkthdr *>(&packet.data().header()),
                         packet.data().begin());
        }
    }

    void
    analyzer_type::analyze_packet(network_packet const &packet) {
        if (_first_packet) {
            _first_packet = false;
            _first_packet_timestamp = packet.data().timestamp();
        } else if (packet.data().timestamp() < _first_packet_timestamp) {
            cerr << "Packet timestamp before first packet, skipping....\n";
            return;
        }

        timeval timestamp = (packet.data().timestamp()
                             - _first_packet_timestamp);
        current_packet_timestamp = (float) timeval_to_double(timestamp);
        for (set<listener_type *>::const_iterator it = _listeners.begin();
             it != _listeners.end();
             ++it)
        {
            listener_type *listener = *it;
            listener->analyze(*this, timestamp, packet);
        }
    }

    void
    analyzer_type::analyze_packets() {
        pcap_pkthdr header;
        while (u_char const *data = _packets.next(&header)) {
            ethernet_packet packet(packet_data(data, header));
            try {
                analyze_packet(packet);
                ++_packet_count;
            } catch (runtime_error &ex) {
                display_exception("runtime_error", ex, packet);
            } catch (out_of_range &ex) {
                display_exception("out_of_range", ex, packet);
            } catch (exception &ex) {
                display_exception("exception", ex, packet);
            }
        }
    }

    bool
    analyzer_type::analyze_trace_files() {
        _packet_count = 0;
        _first_packet = true;
        _first_packet_timestamp.tv_sec = 0;
        _first_packet_timestamp.tv_usec = 0;

        for (set<listener_type *>::const_iterator it = _listeners.begin();
             it != _listeners.end();
             ++it)
        {
            listener_type *listener = *it;
            listener->initialize(*this);
        }

        for (set<string>::const_iterator it = _trace_filenames.begin();
             it != _trace_filenames.end();
             ++it)
        {
            string const &filename = *it;
            _packets.open_offline(filename);
            if (!_packets.good()) {
                cerr << _program_name << ": unable to open '"
                     << filename << "': " << _packets.error() << "\n";
                return false;
            }
            if (!_dump_interesting.empty()) {
                _dumper.open(_dump_interesting);
            }
            if (!_dumper.good()) {
                cerr << _program_name << ": unable to open '"
                     << _dump_interesting << "': " << _dumper.error() << '\n';
                return false;
            }

            analyze_packets();

            _dumper.close();
            _packets.close();
        }

        for (set<listener_type *>::const_iterator it = _listeners.begin();
             it != _listeners.end();
             ++it)
        {
            listener_type *listener = *it;
            listener->finalize(*this);
        }

        return true;
    }
}
