// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef ANALYZER_HH
#define ANALYZER_HH

#include "pcap.hh"

#include <fstream>
#include <set>
#include <stdexcept>
#include <string>
#include <vector>

namespace dns_analyzer {

    using std::exception;
    using std::ofstream;
    using std::set;
    using std::string;
    using std::vector;

    class analyzer_type;
    class network_packet;

    /**
     * The timestamp (in seconds) of the current packet relative to
     * the first packet.  This is only here because it is needed by
     * the default constructor of interval_stat_type.
     */
    extern float current_packet_timestamp;

    /**
     * Interface defined for packet analyzer modules. Every analysis
     * must implement this interface and register with the analyzer.
     */
    class listener_type {
    public:
        virtual ~listener_type();

        /**
         * Initialize is invoked by the analyzer just before reading
         * the first packet.
         *
         * @param analyzer the analyzer reading the packets.
         */
        virtual void initialize(analyzer_type &analyzer) = 0;

        /**
         * Analyze is invoked once for each packet read by the
         * analyzer.
         *
         * @param analyzer the analyzer reading the packets.
         *
         * @param timestamp the timestamp of the packet relative to
         * the timestamp of the first packet.
         *
         * @param packet the packet.
         */
        virtual void analyze(analyzer_type &analyzer,
                             timeval const &timestamp,
                             network_packet const &packet) = 0;

        /**
         * Finalize is invoked after the last packet has been read by
         * the analyzer.
         *
         * @param analyzer the analyzer reading the packets.
         */
        virtual void finalize(analyzer_type &analyzer) = 0;
    };

    /**
     * The analyzer responsible for reading network packets and
     * passing them to the actual analysis modules.  The analyzer is
     * also responsible for keeping track of packets that are
     * considered "interesting" and some other administrative data.
     */
    class analyzer_type {
    public:
        explicit analyzer_type(string const &program_name);

        /**
         * Enable or disable the anonymization of packet data (such as
         * domain names and IP addresses).
         *
         * @param value enable status.
         */
        void anonymize(bool value) { _anonymize = value; }

        /**
         * Is anonymizing of packet data enabled?
         */
        bool anonymize() const { return _anonymize; }

        /**
         * Enable or disable gathering of interesting packets.
         *
         * @param value enable status.
         */
        void interesting(bool value) { _interesting = value; }

        /**
         * Is gathering of interesting packets enabled?
         */
        bool interesting() const { return _interesting; }

        /**
         * Set the name of the file used to write interesting packets
         * to disk.  This file is in the standard tcpdump format.  If
         * the filename is empty the packets are not dumped.
         *
         * @param value valid file name or empty string.
         */
        void dump_interesting(string const &value) { _dump_interesting = value; }

        /**
         * The name of the file used to write interesting packets to
         * disk.
         */
        string const &dump_interesting() const { return _dump_interesting; }

        /**
         * Add an additional file containing packet traces.  Every
         * file is read and analyzed in order of insertion.  Duplicate
         * filenames are removed.
         *
         * @param filename valid file name.
         */
        void insert_trace_filename(string const &filename);

        /**
         * Add an analysis module to the analyzer.  Duplicate analysis
         * modules are ignored.
         *
         * @param listener non-null pointer to listener [object].
         */
        void insert_listener(listener_type *listener);

        /**
         * Remove an analysis module from the analyzer.
         *
         * @param listener non-null pointer to listener [function].
         */
        void erase_listener(listener_type *listener);

        /**
         * Stores a deep copy of 'packet' in the list of interesting
         * packets if the gathering of interesting packets is enabled.
         *
         * @param packet packet to be copied and added to the
         * interesting packet list [function].
         */
        void insert_interesting(network_packet const &packet);

        /**
         * Give the list of interesting packets gathered during
         * analysis.
         *
         * @return list of interesting packets [object].
         */
        vector<network_packet *> const &interesting_packets() const {
            return _interesting_packets;
        }

        /**
         * Read all the packets from the indicated trace file or
         * device and have the analysis modules process them.
         *
         * @return success indicator.
         */
        bool analyze_trace_files();

        /**
         * The number of packets analyzed so far.
         */
        size_t packet_count() const { return _packet_count; }

    private:
        packet_capture _packets;
        packet_dumper_type _dumper;
        string _program_name;
        bool _anonymize;
        bool _interesting;
        string _dump_interesting;
        set<string> _trace_filenames;
        set<listener_type *> _listeners;
        size_t _packet_count;
        bool _first_packet;
        timeval _first_packet_timestamp;
        vector<network_packet *> _interesting_packets;

        void analyze_packets();
        void analyze_packet(network_packet const &packet);
        void display_exception(char const *msg, exception const &ex, network_packet const &packet);
    };
}

#endif // ANALYZER_HH
