// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "analyzer.hh"
#include "aggregation.hh"
#include "packet.hh"
#include "performance.hh"
#include "pointers.hh"
#include "r-project.hh"
#include "variable.hh"

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <unistd.h>
#include <cstring>

using namespace dns_analyzer;

using std::cerr;
using std::cout;
using std::invalid_argument;
using std::istringstream;
using std::strcmp;

static char const *program_name;

void
usage() {
    cerr << "Usage: " << program_name << " [options] tcpdumpfile...\n"
         << "DNS Analyzer version " PACKAGE_VERSION ".\n\n"
         << "Available options:\n"
         << "    -?, -h         Print this usage information.\n"
         << "    -A             Print all aggregated information.\n"
         << "    -a             Anonymize R output.\n"
         << "    -d filename    Dump \"interesting\" packets to filename.\n"
         << "    -g fields      Gather aggregated information on fields.\n"
         << "    -i             Display \"interesting\" packets.\n"
         << "    -m count       Split R output on count records.\n"
         << "    -p count       Prune statistics every count packets.\n"
         << "    -q             Only aggregate queries.\n"
         << "    -r             Only aggregate responses.\n"
         << "    -T intervals   Specify intervals for measuring throughput.\n"
         << "    -t count       Dump the top count aggregated data.\n"
         << "    -w filename    Write packets to file in R data format.\n"
         << "    -s             Display number of packets parsed per second.\n"
         << "    --             Ignore all further options.\n\n"
         << "Report bugs to <" PACKAGE_BUGREPORT ">.\n";
}

/**
 * Main program.
 */
int
main(int argc, char **argv) {
    program_name = argv[0];

    if (argc < 2) {
        usage();
        return EXIT_FAILURE;
    }

    analyzer_type analyzer(program_name);

    size_t current_prune_count = 0;
    size_t current_top_count = 0;
    bool current_print_all = false;
    bool current_handle_responses = true;
    bool current_handle_queries = !current_handle_responses;
    scoped_ptr<r_project_listener_type> r_project_listener;
    scoped_ptr<performance_listener_type> performance_listener;
    scoped_ptr<combined_stat_config_type> current_interval_times;

    int ch;
    while ((ch = getopt(argc, argv, "Aad:g:him:p:qrsT:t:w:")) != -1) {
        switch (ch) {
        case 'A':
            current_print_all = true;
            break;
        case 'a':
            analyzer.anonymize(true);
            break;
        case 'd':
            analyzer.dump_interesting(optarg);
            break;
        case 'g': {
            aggregation_listener_type *listener = new aggregation_listener_type;
            try {
                listener->set_fields(make_field_list(optarg));
            } catch (invalid_argument &ex) {
                cerr << program_name << ": invalid field name '"
                     << ex.what() << "'\n";
                return EXIT_FAILURE;
            }
            listener->set_interval_times(current_interval_times);
            listener->prune_count(current_prune_count);
            listener->top_count(current_top_count);
            listener->print_all(current_print_all);
            listener->handle_responses(current_handle_responses);
            listener->handle_queries(current_handle_queries);
            analyzer.insert_listener(listener);
            break;
        }
        case 'i':
            analyzer.interesting(true);
            break;
        case 'm': {
            istringstream is(optarg);
            size_t count;
            is >> count;

            if (!r_project_listener) {
                r_project_listener.reset(new r_project_listener_type);
            }
            r_project_listener->split_count(count);
            break;
        }
        case 'p': {
            istringstream is(optarg);
            is >> current_prune_count;
            break;
        }
        case 'q':
            current_handle_queries = true;
            current_handle_responses = false;
            break;
        case 'r':
            current_handle_responses = true;
            current_handle_queries = false;
            break;
        case 's':
            if (!performance_listener) {
                performance_listener.reset(new performance_listener_type);
            }
            break;
        case 'T': {
            vector<float> term_intervals;
            float temp;
            istringstream in(optarg);
            while (in >> temp) {
                term_intervals.push_back(temp);
                in.get();           // ','
            }
            if (!term_intervals.empty()) {
                current_interval_times.reset(
                    new combined_stat_config_type(term_intervals));
            } else {
                current_interval_times.reset(0);
            }
            break;
        }
        case 't': {
            istringstream is(optarg);
            is >> current_top_count;
            break;
        }
        case 'w':
            if (!r_project_listener) {
                r_project_listener.reset(new r_project_listener_type);
            }
            r_project_listener->base_filename(optarg);
            break;
        case '?':
        case 'h':
        default:
            usage();
            return EXIT_FAILURE;
        }
    }

    argc -= optind;
    argv += optind;

    for (int i = 0; i < argc; ++i) {
        analyzer.insert_trace_filename(argv[i]);
    }

    if (r_project_listener) {
        analyzer.insert_listener(r_project_listener);
    }
    if (performance_listener) {
        analyzer.insert_listener(performance_listener);
    }

    bool result = analyzer.analyze_trace_files();

    if (!result) {
        return EXIT_FAILURE;
    }

    if (analyzer.interesting()) {
        cerr << "Captured " << analyzer.interesting_packets().size()
             << " strange packets.\n";

        for (vector<network_packet *>::const_iterator it
                 = analyzer.interesting_packets().begin();
             it != analyzer.interesting_packets().end();
             ++it)
        {
            network_packet *p = *it;
            p->print_info(cout);
            delete p;
        }
    }

    return EXIT_SUCCESS;
}
