// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "packet-dns.hh"

#include "value.hh"

#include <algorithm>
#include <cctype>
#include <iosfwd>
#include <stdexcept>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <cstring>

namespace dns_analyzer {

    using std::dec;

    struct ns_rcode_info_type {
        ns_rcode rcode;
        char const *name;
    };

    static ns_rcode_info_type const ns_rcode_table[] = {
        { ns_r_noerror, "NOERROR" },
        { ns_r_formerr, "FORMERR" },
        { ns_r_servfail, "SERVFAIL" },
        { ns_r_nxdomain, "NXDOMAIN" },
        { ns_r_notimpl, "NOTIMPL" },
        { ns_r_refused, "REFUSED" },
        { ns_r_yxdomain, "YXDOMAIN" },
        { ns_r_yxrrset, "YXRRSET" },
        { ns_r_nxrrset, "NXRRSET" },
        { ns_r_notauth, "NOTAUTH" },
        { ns_r_notzone, "NOTZONE" },
        { ns_r_max, 0 },
    };

    struct ns_type_info_type {
        ns_type type;
        char const *name;
        char const *format;
    };

    static ns_type_info_type const ns_type_table[] = {
        { ns_t_a, "A", "4" },
        { ns_t_ns, "NS", "n" },
        { ns_t_md, "MD", "n" },
        { ns_t_mf, "MF", "n" },
        { ns_t_cname, "CNAME", "n" },
        { ns_t_soa, "SOA", "nnlllll" },
        { ns_t_mb, "MB", "n" },
        { ns_t_mg, "MG", "n" },
        { ns_t_mr, "MR", "n" },
        { ns_t_null, "NULL", "" },
        { ns_t_wks, "WKS", "" },
        { ns_t_ptr, "PTR", "n" },
        { ns_t_hinfo, "HINFO", "tt" },
        { ns_t_minfo, "MINFO", "nn" },
        { ns_t_mx, "MX", "sn" },
        { ns_t_txt, "TXT", "t" },
        { ns_t_rp, "RP", "???"},                        // RFC1183
        { ns_t_afsdb, "AFSDB", "???"},                  // RFC1183
        { ns_t_x25, "X25", "???"},                      // RFC1183
        { ns_t_isdn, "ISDN", "???"},                    // RFC1183
        { ns_t_rt, "RT", "???"},                        // RFC1183
        { ns_t_nsap, "NSAP", "???"},                    // RFC1706
        { ns_t_nsap_ptr, "NSAP-PTR", "???"},            // RFC1706
        { ns_t_sig, "SIG", "???"},                      // RFC2535
        { ns_t_key, "KEY", "???"},                      // RFC2535
        { ns_t_px, "PX", "???"},                        // RFC2163
        { ns_t_gpos, "GPOS", "???"},                    // RFC1712
        { ns_t_aaaa, "AAAA", "6" },                     // [Thomson]
        { ns_t_loc, "LOC", "???" },                     // [Vixie]
        { ns_t_nxt, "NXT", "???" },                     // RFC2535
        { ns_t_eid, "EID", "???" },                     // [Patton]
        { ns_t_nimloc, "NIMLOC", "???" },               // [Patton]
        { ns_t_srv, "SRV", "sssn" },                    // RFC2782
        { ns_t_atma, "ATMA", "???" },                   // Dobrowski
        { ns_t_naptr, "NAPTR", "sstttn" },              // RFC2168, RFC2915
        { static_cast<ns_type>(36), "KX", "???" },      // RFC2230
        { static_cast<ns_type>(37), "CERT", "???" },    // RFC2538
        { static_cast<ns_type>(38), "A6", "???" },      // RFC2874
        { static_cast<ns_type>(39), "DNAME", "???" },   // RFC2672
        { static_cast<ns_type>(40), "SINK", "???" },    // Eastlake
        { ns_t_opt, "OPT", "???" },                     // RFC2671
        { static_cast<ns_type>(42), "APL", "???" },     // RFC3123

        { static_cast<ns_type>(100), "UINFO", "???" },  // [IANA-Reserved]
        { static_cast<ns_type>(101), "UID", "???" },    // [IANA-Reserved]
        { static_cast<ns_type>(102), "GID", "???" },    // [IANA-Reserved]
        { static_cast<ns_type>(103), "UNSPEC", "???" }, // [IANA-Reserved]

        { static_cast<ns_type>(249), "TKEY", "???" },   // RFC2930
        { static_cast<ns_type>(250), "TSIG", "???" },   // RFC2845

        { ns_t_ixfr, "IXFR", "???" },                   // RFC1995
        { ns_t_axfr, "AXFR", "???" },
        { ns_t_mailb, "MAILB", "" },
        { ns_t_maila, "MAILA", "" },
        { ns_t_any, "ANY", "" },

        { static_cast<ns_type>(0), 0, 0 }
    };

    static ns_type_info_type const *
    get_ns_type_info(ns_type type) {
        for (ns_type_info_type const *it = ns_type_table; it->type; it++) {
            if (type == it->type) {
                return it;
            }
        }
        return 0;
    }

    struct ns_class_info_type {
        ns_class the_class;
        char const *name;
    };

    static ns_class_info_type const ns_class_table[] = {
        { ns_c_in, "IN" },
        { ns_c_chaos, "CH" },
        { ns_c_hs, "HS" },
        { ns_c_none, "NONE" },
        { ns_c_any, "ANY" },
        { static_cast<ns_class>(0), 0 }
    };

    static ns_class_info_type const *
    get_ns_class_info(ns_class the_class) {
        for (ns_class_info_type const *it = ns_class_table;
             it->the_class;
             it++)
        {
            if (the_class == it->the_class) {
                return it;
            }
        }
        return 0;
    }

    u_short
    get16(u_char const *cp) {
        u_short result;
        NS_GET16(result, cp);
        return result;
    }
    
    u_long
    get32(u_char const *cp) {
        u_long result;
        NS_GET32(result, cp);
        return result;
    }

    static inline size_t
    label_length(u_char const *label) {
        return *label;
    }

    static inline size_t
    label_size(u_char const *label) {
        return *label + 1;
    }

    static inline u_char const *
    label_next(u_char const *label) {
        return *label ? label + *label + 1 : 0;
    }

    static inline u_char const *
    label_start(u_char const *label) {
        return label + 1;
    }

    static inline u_char const *
    label_end(u_char const *label) {
        return label_start(label) + label_length(label);
    }

    static inline bool
    label_equal(u_char const *lhs, u_char const *rhs) {
        return std::equal(label_start(lhs), label_end(lhs), label_start(rhs));
    }

    static inline bool
    is_pointer(u_char const *label) {
        return ((label_length(label) & 0xC0) >> 6) == 3;
    }

    static inline bool
    is_reserved(u_char const *label) {
        u_char bits = (label_length(label) & 0xC0) >> 6;
        return bits == 1 || bits == 2;
    }

    static inline bool
    is_label(u_char const *label) {
        return ((label_length(label) & 0xC0) >> 6) == 0;
    }

    domain_name_type::domain_name_type(packet_data const &data, size_t dname_start)
        : _data(data), _dname_start(dname_start)
    {
    }

    domain_name_type::domain_name_type(domain_name_type const &rhs)
        : _data(rhs._data), _dname_start(rhs._dname_start)
    {
    }

    domain_name_type const &
    domain_name_type::operator=(domain_name_type const &rhs) {
        if (this != &rhs) {
            _data = rhs._data;
            _dname_start = rhs._dname_start;
        }
        return *this;
    }

    bool
    domain_name_type::operator==(domain_name_type const &rhs) const {
        if (this == &rhs) {
            return true;
        }

        domain_name_type const &lhs = *this;

        u_char const *label_1 = &lhs._data.at(_dname_start);
        u_char const *label_2 = &rhs._data.at(_dname_start);
        while (label_length(label_1) == label_length(label_2)) {
            if (label_length(label_1) == 0) {
                return true;
            }
            if (!label_equal(label_1, label_2)) {
                return false;
            }
            label_1 = lhs.next_label(label_1);
            label_2 = rhs.next_label(label_2);
        }
        return false;
    }

    bool
    domain_name_type::operator!=(domain_name_type const &rhs) const {
        return !operator==(rhs);
    }

    bool
    domain_name_type::operator< (domain_name_type const &rhs) const {
        return this->to_string() < rhs.to_string();
    }

    bool
    domain_name_type::operator<=(domain_name_type const &rhs) const {
        return this->to_string() <= rhs.to_string();
    }

    bool
    domain_name_type::operator> (domain_name_type const &rhs) const {
        return this->to_string() > rhs.to_string();
    }

    bool
    domain_name_type::operator>=(domain_name_type const &rhs) const {
        return this->to_string() >= rhs.to_string();
    }

    size_t
    domain_name_type::compressed_size() const {
        size_t result = 0;
        u_char const *label = &_data.at(_dname_start);
        while (is_label(label) && label_length(label) > 0) {
            result += label_size(label);
            label = label_end(label);
        }
        if (is_label(label)) {
            // Terminated by origin.
            result += 1;
        } else if (is_pointer(label)) {
            // Terminated by compressed label pointer.
            result += 2;
        } else {
            throw std::runtime_error("unknown label type");
        }
        return result;
    }

    size_t
    domain_name_type::size() const {
        size_t result = 1;
        u_char const *label = follow_pointer(&_data.at(_dname_start));
        while (label_length(label) > 0) {
            result += label_size(label);
            label = next_label(label);
        }
        return result;
    }

    string
    domain_name_type::to_string() const {
        string result;

        u_char const *label = follow_pointer(&_data.at(_dname_start));
        if (*label) {
            while (*label) {
                result.append(reinterpret_cast<char const *>(label + 1),
                              *label);
                label = next_label(label);
                if (*label) {
                    result += '.';
                }
            }
        } else {
            result = ".";
        }

        // Lowercase the result.
        std::transform(result.begin(), result.end(),
                       result.begin(), (int (*)(int)) std::tolower);
        return result;
    }

    u_char const *
    domain_name_type::follow_pointer(u_char const *label) const {
        // See RFC1035 section 4.1.4.
        if (label == 0) {
            return 0;
        }

        switch ((label[0] & 0xC0) >> 6) {
        case 0:
            return label;
        case 3: {
            u_short offset = ((label[0] & 0x3F) << 8) | label[1];
            return follow_pointer(&_data.at(offset));
        }
        case 1:
        case 2:
        default:
            throw std::runtime_error("invalid dname pointer (reserved bits)");
        }
    }

    u_char const *
    domain_name_type::next_label(u_char const *label) const {
        return follow_pointer(label_next(label));
    }

    ostream &
    operator<<(ostream &out, domain_name_type const &rhs) {
        return out << rhs.to_string();
    }


    // resource_record_type

    resource_record_type::resource_record_type(dns_packet const *packet, size_t offset,
                                     section_type section)
        : _packet(packet), _offset(offset), _section(section)
    {
    }

    bool
    resource_record_type::is_question() const {
        return _section == question_section;
    }

    section_type
    resource_record_type::section() const {
        return _section;
    }

    domain_name_type
    resource_record_type::name() const {
        return domain_name_type(_packet->data(), _offset);
    }

    ns_type
    resource_record_type::type() const {
        return static_cast<ns_type>(
            get16(&_packet->data().at(_offset + name_size())));
    }

    ns_class
    resource_record_type::the_class() const {
        return static_cast<ns_class>(
            get16(&_packet->data().at(_offset + name_size() + 2)));
    }

    u_long
    resource_record_type::ttl() const {
        return get32(&_packet->data().at(_offset + name_size() + 4));
    }

    u_short
    resource_record_type::rdata_length() const {
        return get16(&_packet->data().at(_offset + name_size() + 8));
    }

    size_t
    resource_record_type::byte_size() const {
        return name_size() + (is_question() ? 4 : 10 + rdata_length());
    }

    size_t
    resource_record_type::size() const {
        if (is_question()) {
            return 0;
        } else if (ns_type_info_type const *info = get_ns_type_info(type())) {
            return strlen(info->format);
        } else {
            return (size_t) -1;
        }
    }

    resource_record_type::const_iterator
    resource_record_type::begin() const {
        return const_iterator(this, _offset + rdata_start_offset());
    }

    resource_record_type::const_iterator
    resource_record_type::end() const {
        return const_iterator(this, _offset + byte_size());
    }

    size_t
    resource_record_type::rdata_start_offset() const {
        return name_size() + (is_question() ? 4 : 10);
    }

    size_t
    resource_record_type::name_size() const {
        return domain_name_type(_packet->data(), _offset).compressed_size();
    }


    // resource_record_type::const_iterator

    resource_record_type::const_iterator::const_iterator(
        resource_record_type const *rr, size_t offset)
        : _rr(rr),
          _index(0),
          _offset(offset)
    {
        ns_type_info_type const *info = get_ns_type_info(rr->type());
        _format = (info
                   ? (info->format ? info->format : "r")
                   : "r");
    }

    resource_record_type::const_iterator::const_iterator(
        const_iterator const &rhs)
        : _rr(rhs._rr),
          _format(rhs._format),
          _index(rhs._index),
          _offset(rhs._offset)
    {
    }

    resource_record_type::const_iterator const &
    resource_record_type::const_iterator::operator=(const_iterator const &rhs) {
        if (this != &rhs) {
            _rr = rhs._rr;
            _format = rhs._format;
            _index = rhs._index;
            _offset = rhs._offset;
        }

        return *this;
    }

    bool
    resource_record_type::const_iterator::operator==(const_iterator const &rhs) const {
        if (this == &rhs) {
            return true;
        }

        return _rr == rhs._rr && _offset == rhs._offset;
    }

    bool
    resource_record_type::const_iterator::operator!=(const_iterator const &rhs) const {
        return !operator==(rhs);
    }

    value_ref
    resource_record_type::const_iterator::operator*() const {
        u_char const *data = &_rr->_packet->data()[_offset];
        switch (_format[_index]) {
        case 0:
            throw std::logic_error("end iterator dereferenced");
        case '4':
            return value_ref(
                typed_value<internet_address>(
                    internet_address(* (in_addr *) data)));
        case '6':
            return value_ref(
                typed_value<internet_address>(
                    internet_address(* (in6_addr *) data)));
        case 'l':
            return value_ref(typed_value<u_long>(get32(data)));
        case 'n':
            return value_ref(typed_value<domain_name_type>(
                                 domain_name_type(_rr->_packet->data(), _offset)));
        case 's':
            return value_ref(typed_value<u_short>(get16(data)));
        case 't':
            return value_ref(typed_value<string>(string((char *) data + 1,
                                                        *data)));
        case 'r':
            throw std::logic_error("cannot reference raw data");
        default:
            throw std::runtime_error("unknown rdata type");
        }
    }

    resource_record_type::const_iterator const &
    resource_record_type::const_iterator::operator++() {
        switch (_format[_index]) {
        case 0:
            throw std::logic_error("end iterator incremented");
        case '4':
            _offset += sizeof(in_addr);
            break;
        case '6':
            _offset += sizeof(in6_addr);
            break;
        case 'l':
            _offset += sizeof(u_long);
            break;
        case 'n':
            _offset += (domain_name_type(_rr->_packet->data(), _offset)
                        .compressed_size());
            break;
        case 's':
            _offset += sizeof(u_short);
            break;
        case 't':
            // Text string.
            _offset += 1 + _rr->_packet->data()[_offset];
            break;
        case 'r':
            // Raw (unknown) data.
            _offset = _rr->_offset + _rr->byte_size();
            break;
        default:
            throw std::runtime_error("unknown rdata type");
        }
        ++_index;
        return *this;
    }

    resource_record_type::const_iterator const
    resource_record_type::const_iterator::operator++(int) {
        const_iterator result(*this);
        operator++();
        return result;
    }


    // dns_packet

    static inline HEADER const *
    dns_packet_header(dns_packet const *packet) {
        return reinterpret_cast<HEADER const *>(
            packet->data().begin() + (packet->is_tcp() ? sizeof(u_short) : 0));
    }

    dns_packet::dns_packet(packet_data const &data,
                           network_packet const *container, size_t tcp_length)
        : network_packet(data, container),
          _dns_data(data.slice(tcp_length ? sizeof(u_short) : 0,
                               data.captured_size())),
          _tcp_length(tcp_length)
    {
    }

    dns_packet::~dns_packet() {
    }

    void
    dns_packet::print_payload_info(ostream &out, size_t indent) const {
        out << dec << "dns" << (is_query() ? "?" : "!")
            << " qdcount: " << question_count()
            << ", ancount: " << answer_count()
            << ", nscount: " << authority_count()
            << ", arcount: " << additional_count()
            << "\n";
        for (dns_packet::const_iterator it = begin(); it != end(); ++it)
        {
            resource_record_type rr(*it);
            out << string(indent, ' ')
                << rr.name().to_string()
                << " " << rr.the_class()
                << " " << rr.type();
            for (resource_record_type::const_iterator it2 = rr.begin();
                 it2 != rr.end();
                 ++it2)
            {
                out << " " << *it2;
            }

            out << "\n";
        }
    }

    dns_packet *
    dns_packet::shallow_copy() const {
        if (container() != 0) {
            throw std::logic_error("cannot copy payload packet");
        }
        return new dns_packet(data());
    }

    bool
    dns_packet::good() const {
        return (data().captured_size() >= sizeof(HEADER)
                && data().captured_size() <= std::max((size_t) NS_PACKETSZ, _tcp_length));
    }

    size_t
    dns_packet::header_size() const {
        return sizeof(HEADER) + (_tcp_length > 0 ? sizeof(u_short) : 0);
    }

    network_packet const *
    dns_packet::payload() const {
        return 0;
    }

    u_short
    dns_packet::id() const {
        return ntohs(dns_packet_header(this)->id);
    }

    bool
    dns_packet::is_query() const {
        return dns_packet_header(this)->qr == 0;
    }

    ns_opcode
    dns_packet::opcode() const {
        return static_cast<ns_opcode>(dns_packet_header(this)->opcode);
    }

    bool
    dns_packet::is_authoritative_answer() const {
        return dns_packet_header(this)->aa;
    }

    bool
    dns_packet::is_truncated() const {
        return dns_packet_header(this)->tc;
    }

    bool
    dns_packet::is_recursion_desired() const {
        return dns_packet_header(this)->rd;
    }

    bool
    dns_packet::is_recursion_available() const {
        return dns_packet_header(this)->ra;
    }

    bool
    dns_packet::is_authentic_data() const {
        return dns_packet_header(this)->ad;
    }

    bool
    dns_packet::is_checking_disabled() const {
        return dns_packet_header(this)->cd;
    }

    ns_rcode
    dns_packet::rcode() const {
        return static_cast<ns_rcode>(dns_packet_header(this)->rcode);
    }

    u_short
    dns_packet::section_count(section_type section) const {
        switch (section) {
        case question_section:
            return question_count();
        case answer_section:
            return answer_count();
        case authority_section:
            return authority_count();
        case additional_section:
            return additional_count();
        case no_section:
            return 0;
        default:
            throw std::logic_error("section_count");
        }
    }

    u_short
    dns_packet::question_count() const {
        return ntohs(dns_packet_header(this)->qdcount);
    }

    u_short
    dns_packet::answer_count() const {
        return ntohs(dns_packet_header(this)->ancount);
    }

    u_short
    dns_packet::authority_count() const {
        return ntohs(dns_packet_header(this)->nscount);
    }

    u_short
    dns_packet::additional_count() const {
        return ntohs(dns_packet_header(this)->arcount);
    }

    dns_packet::const_iterator
    dns_packet::begin() const {
        return const_iterator(this, header_size(), question_section);
    }

    dns_packet::const_iterator
    dns_packet::end() const {
        return const_iterator(this, data().captured_size(), no_section);
    }

    bool
    dns_packet::is_tcp() const {
        return _tcp_length > 0;
    }

    value_ref
    dns_packet::field_qcount() const {
        return typed_value<u_short>(question_count());
    }

    value_ref
    dns_packet::field_type() const {
        if (question_count() >= 1) {
            resource_record_type question(*begin());
            return new typed_value<ns_type>(question.type());
        } else {
            return null_value;
        }
    }

    value_ref
    dns_packet::field_class() const {
        if (question_count() >= 1) {
            resource_record_type question(*begin());
            return new typed_value<ns_class>(question.the_class());
        } else {
            return null_value;
        }
    }

    value_ref
    dns_packet::field_domain() const {
        if (question_count() >= 1) {
            resource_record_type question(*begin());
            return new typed_value<string>(question.name().to_string());
        } else {
            return null_value;
        }
    }

    value_ref
    dns_packet::field_query() const {
        return new typed_value<bool>(is_query());
    }

    value_ref
    dns_packet::field_authoritative() const {
        return new typed_value<bool>(is_authoritative_answer());
    }

    value_ref
    dns_packet::field_truncated() const {
        return new typed_value<bool>(is_truncated());
    }

    value_ref
    dns_packet::field_rcode() const {
        return new typed_value<ns_rcode>(rcode());
    }

    value_ref
    dns_packet::field_client() const {
        if (is_query()) {
            return new typed_value<internet_address>(
                find_up<ip_packet>(this)->source_address());
        } else {
            return new typed_value<internet_address>(
                find_up<ip_packet>(this)->dest_address());
        }
    }

    value_ref
    dns_packet::field_server() const {
        if (is_query()) {
            return new typed_value<internet_address>(
                find_up<ip_packet>(this)->dest_address());
        } else {
            return new typed_value<internet_address>(
                find_up<ip_packet>(this)->source_address());
        }
    }


    // dns_packet::const_iterator
    dns_packet::const_iterator::const_iterator(
        dns_packet const *packet, size_t offset, section_type section)
        : _packet(packet),
          _offset(offset),
          _section(section),
          _remaining_count(packet->section_count(section))
    {
    }

    dns_packet::const_iterator::const_iterator(const_iterator const &rhs)
        : _packet(rhs._packet),
          _offset(rhs._offset),
          _section(rhs._section),
          _remaining_count(rhs._remaining_count)
    {
    }

    dns_packet::const_iterator const &
    dns_packet::const_iterator::operator=(const_iterator const &rhs) {
        if (this != &rhs) {
            _packet = rhs._packet;
            _offset = rhs._offset;
            _section = rhs._section;
            _remaining_count = rhs._remaining_count;
        }

        return *this;
    }

    bool
    dns_packet::const_iterator::operator==(const_iterator const &rhs) const {
        return _packet == rhs._packet && _offset == rhs._offset;
    }

    bool
    dns_packet::const_iterator::operator!=(const_iterator const &rhs) const {
        return !operator==(rhs);
    }

    resource_record_type
    dns_packet::const_iterator::operator*() const {
        return resource_record_type(_packet, _offset, _section);
    }

    dns_packet::const_iterator const &
    dns_packet::const_iterator::operator++() {
        _offset += operator*().byte_size();
        --_remaining_count;
        if (_remaining_count == 0) {
            _section = static_cast<section_type>(_section + 1);
            _remaining_count = _packet->section_count(_section);
        }
        return *this;
    }

    dns_packet::const_iterator const
    dns_packet::const_iterator::operator++(int) {
        const_iterator result(*this);
        operator++();
        return result;
    }


    ostream &
    operator<<(ostream &out, ns_rcode rcode) {
        if (rcode < ns_r_noerror || rcode >= ns_r_max) {
            out << "UNKNOWN";
        } else {
            out << ns_rcode_table[rcode].name;
        }
        return out;
    }

    ostream &
    operator<<(ostream &out, ns_type type) {
        ns_type_info_type const *info = get_ns_type_info(type);
        if (info) {
            out << info->name;
        } else {
            out << static_cast<int>(type);
        }
        return out;
    }

    ostream &
    operator<<(ostream &out, ns_class the_class) {
        ns_class_info_type const *info = get_ns_class_info(the_class);
        if (info) {
            out << info->name;
        } else {
            out << static_cast<int>(the_class);
        }
        return out;
    }
}
