// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef PACKET_DNS_HH
#define PACKET_DNS_HH

#include "fast_map.hh"
#include "packet.hh"
#include "util.hh"

#include <arpa/nameser.h>

namespace dns_analyzer {

    /**
     * A domain_name is a sequence of domain_labels terminated by a
     * zero-length domain_label.  The maximum size of a domain_name is
     * 255 octets (including the length octets of the labels).
     *
     * This class uses the complete DNS packet data to interpret
     * compressed dnames correctly.
     */
    class domain_name_type {
    public:
        /**
         * Construct a new domain name.
         *
         * @param data the packet data containing the domain name
         * [object].
         *
         * @param dname_start the offset where the domain name starts.
         */
        domain_name_type(packet_data const &data, size_t dname_start);

        /**
         * Copy constructor.
         */
        domain_name_type(domain_name_type const &rhs);

        /**
         * Assignment.
         */
        domain_name_type const &operator=(domain_name_type const &rhs);

        bool operator==(domain_name_type const &rhs) const;
        bool operator!=(domain_name_type const &rhs) const;

        bool operator< (domain_name_type const &rhs) const;
        bool operator<=(domain_name_type const &rhs) const;
        bool operator> (domain_name_type const &rhs) const;
        bool operator>=(domain_name_type const &rhs) const;

        /**
         * The total size of the domain name up to (and including) the
         * first pointer.
         */
        size_t compressed_size() const;

        /**
         * The total size of the domain name.
         */
        size_t size() const;

        /**
         * The domain name as a string using the normal dotted
         * notation (for example: "www.w3c.org").  The origin is
         * represented as ".".
         */
        string to_string() const;
    private:
        packet_data _data;
        size_t _dname_start;

        u_char const *follow_pointer(u_char const *label) const;
        u_char const *next_label(u_char const *label) const;
    };

    /**
     * Write the text representation of a domain_name_type instance to
     * the specified ostream.
     */
    ostream &operator<<(ostream &out, domain_name_type const &rhs);


    /**
     * The different resource record data sections in a DNS packet.
     */
    enum section_type {
        // Do _not_ change the order of these fields.
        question_section = 0,
        answer_section = 1,
        authority_section = 2,
        additional_section = 3,
        no_section = 4
    };


    /**
     * A DNS resource record.
     */
    class resource_record_type {
    public:
        class const_iterator;

        /**
         * Construct a new resource record.
         *
         * @param packet the packet containing the resource record
         * [object].
         *
         * @param offset the starting offset of the resource record in
         * the packet data.
         *
         * @param section the section of the resource record.
         */
        resource_record_type(dns_packet const *packet,
                             size_t offset,
                             section_type section);

        /**
         * The total size (in bytes) of the resource record.
         */
        size_t byte_size() const;

        /**
         * Is this resource record in the question section?  In this
         * case no ttl or rdata is present for the resource record.
         */
        bool is_question() const;

        /**
         * The section the resource record is located in.
         */
        section_type section() const;

        /**
         * The domain name of the resource record.
         */
        domain_name_type name() const;

        /**
         * The DNS type of the resource record.
         */
        ns_type type() const;

        /**
         * The DNS class of the resource record.
         */
        ns_class the_class() const;

        /**
         * The TTL of the resource record.  This data is not present
         * if the resource record is in the question section of the
         * DNS packet.
         */
        u_long ttl() const;

        /**
         * The length of the rdata of the resource record.  This data
         * is not present if the resource record is in the question
         * section of the DNS packet.
         */
        u_short rdata_length() const;

        /**
         * The rdata of the resource record.  This data is not present
         * if the resource record is in the question section of the
         * DNS packet.
         */
        packet_data rdata() const;

        /**
         * The number of different data elements in the rdata section.
         * This data is not present if the resource record is in the
         * question section of the DNS packet.
         */
        size_t size() const;

        /**
         * The begin iterator of the rdata elements.
         */
        const_iterator begin() const;

        /**
         * The end iterator of the rdata elements.
         */
        const_iterator end() const;


        /**
         * Iterator class for iterating over the rdata of a resource
         * record.
         */
        class const_iterator {
        public:
            /**
             * Copy constructor.
             */
            const_iterator(const_iterator const &rhs);

            /**
             * Assignment.
             */
            const_iterator const &operator=(const_iterator const &rhs);

            /**
             * Compare for equality.
             */
            bool operator==(const_iterator const &rhs) const;

            /**
             * Compare for inequality.
             */
            bool operator!=(const_iterator const &rhs) const;

            /**
             * The current rdata value as a dynamically typed
             * value_ref.  The actual typed pointed to depends on the
             * rdata field.
             */
            value_ref operator*() const;

            /**
             * Advance the iterator to the next element
             * (pre-increment).
             */
            const_iterator const &operator++();

            /**
             * Advance the iterator to the next element
             * (post-increment).
             */
            const_iterator const operator++(int);

        private:
            resource_record_type const *_rr;
            char const *_format;
            size_t _index;
            size_t _offset;

            friend class resource_record_type;
            const_iterator(resource_record_type const *rr, size_t offset);
        };


    private:
        dns_packet const *_packet;
        size_t _offset;
        section_type _section;

        size_t rdata_start_offset() const;
        size_t name_size() const;
    };

    /**
     * This class represents a DNS packet.  The packet can be either a
     * query or a response and can be contained in a UDP packet or a
     * TCP stream.
     */
    class dns_packet : public network_packet {
    public:
        class const_iterator;

        /**
         * Constructor.
         *
         * @param data the data of the IP packet.
         *
         * @param container the containing packet (if any).
         *
         * @param tcp_length the two byte length field (host
         * byte-order) of a DNS packet send over TCP.  The length
         * should be 0 if this packet was send over UDP.
         */
        dns_packet(packet_data const &data,
                   network_packet const *container = 0, size_t tcp_length = 0);

        /**
         * Virtual destructor.
         */
        virtual ~dns_packet();

        virtual void print_payload_info(ostream &out, size_t indent = 0) const;

        virtual dns_packet *shallow_copy() const;

        virtual bool good() const;
        virtual size_t header_size() const;
        virtual network_packet const *payload() const;

        /*
         * The DNS id field.
         */
        u_short id() const;

        /**
         * The DNS opcode field.
         */
        ns_opcode opcode() const;

        /**
         * Is this a DNS query?  Otherwise it is a response.
         */
        bool is_query() const;

        /**
         * The DNS AA field.
         */
        bool is_authoritative_answer() const;

        /**
         * The DNS TC field.
         */
        bool is_truncated() const;

        /**
         * The DNS RD field.
         */
        bool is_recursion_desired() const;

        /**
         * The DNS RA field.
         */
        bool is_recursion_available() const;

        /**
         * The DNSSEC authentic data field.
         */
        bool is_authentic_data() const;

        /**
         * The DNSSEC checking disabled field.
         */
        bool is_checking_disabled() const;

        /**
         * The DNS response code field.
         */
        ns_rcode rcode() const;

        /**
         * The number of entries in the specified section.
         *
         * @param section the section.
         */
        u_short section_count(section_type section) const;

        /**
         * The number of entries in the question section.
         */
        u_short question_count() const;

        /**
         * The number of entries in the answer section.
         */
        u_short answer_count() const;

        /**
         * The number of entries in the authority section.
         */
        u_short authority_count() const;

        /**
         * The number of entries in the additional section.
         */
        u_short additional_count() const;

        /**
         * The forward iterator that points to the first RR in the
         * question section.
         */
        const_iterator begin() const;

        /**
         * The iterator that points just beyond the last RR in the
         * additional section.
         */
        const_iterator end() const;

        /**
         * Was this DNS query/response transferred over TCP?
         */
        bool is_tcp() const;

        /**
         * The question count field as a value_ref.
         */
        value_ref field_qcount() const;

        /**
         * The type field of the first resource record in the question
         * section as a value_ref.
         */
        value_ref field_type() const;

        /**
         * The class field of the first resource record in the
         * question section as a value_ref.
         */
        value_ref field_class() const;

        /**
         * The name field of the first resource record in the question
         * section as a value_ref.
         */
        value_ref field_domain() const;

        /**
         * The query field as a value_ref.
         */
        value_ref field_query() const;

        /**
         * The authoritative field as a value_ref.
         */
        value_ref field_authoritative() const;

        /**
         * The truncated field as a value_ref.
         */
        value_ref field_truncated() const;

        /**
         * The result code field as a value_ref.
         */
        value_ref field_rcode() const;

        /**
         * The internet address of the DNS server as a value_ref.  The
         * server address is the destination address if this packet is
         * a query and the source address if this packet is a
         * response.
         */
        value_ref field_server() const;

        /**
         * The internet address of the DNS clinet as a value_ref.  The
         * client address is the destination address if this packet is
         * a response and the source address if this packet is a
         * query.
         */
        value_ref field_client() const;

        /**
         * This is a forward iterator used to iterate over all the
         * resource records in a DNS packet.
         */
        class const_iterator {
        public:
            const_iterator(const_iterator const &rhs);

            const_iterator const &operator=(const_iterator const &rhs);

            bool operator==(const_iterator const &rhs) const;
            bool operator!=(const_iterator const &rhs) const;

            resource_record_type operator*() const;
            const_iterator const &operator++();
            const_iterator const operator++(int);
        private:
            dns_packet const *_packet;
            size_t _offset;
            section_type _section;  // current section
            u_short _remaining_count; // # of RRs remaining in current section.

            const_iterator(dns_packet const *packet, size_t offset,
                           section_type section);

            friend class dns_packet;
        };

    private:
        packet_data _dns_data;
        size_t _tcp_length;
    };

    /**
     * Write a string representation of a ns_rcode to an ostream.
     */
    ostream &operator<<(ostream &out, ns_rcode rhs);

    /**
     * Write a string representation of a ns_type to an ostream.
     */
    ostream &operator<<(ostream &out, ns_type rhs);

    /**
     * Write a string representation of a ns_class to an ostream.
     */
    ostream &operator<<(ostream &out, ns_class rhs);
}

#ifdef HAVE_EXT_HASH_MAP
namespace __gnu_cxx {
    using namespace dns_analyzer;

    /**
     * The hash of a ns_type.
     */
    template<> struct hash<ns_type> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(ns_type value) const {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a ns_class.
     */
    template<> struct hash<ns_class> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(ns_class value) const {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a ns_rcode.
     */
    template<> struct hash<ns_rcode> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(ns_rcode value) const {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a domain_name_type.
     */
    template<> struct hash<domain_name_type> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(domain_name_type const &value) const {
            hash<std::string> hasher;
            return hasher(value.to_string());
        }
    };
}
#endif // HAVE_EXT_HASH_MAP

#endif // PACKET_DNS_HH
