// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "packet.hh"

#include "value.hh"

#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <ostream>
#include <stdexcept>
#include <sstream>
#include <cstring>

#include <arpa/inet.h>
#include <arpa/nameser.h>
#if defined(HAVE_NET_ETHERNET_H)
# include <net/ethernet.h>
#elif defined(HAVE_SYS_ETHERNET_H)
# include <sys/ethernet.h>
#endif
#ifdef HAVE_NETINET_IN_SYSTM_H
# include <netinet/in_systm.h>
#endif
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/udp.h>

#include "packet-dns.hh"

#if !defined(IPV6_VERSION_MASK)
#define IPV6_VERSION_MASK  0xf0
#endif
#if !defined(IPV6_VERSION)
#define IPV6_VERSION 0x60
#endif

#if BSD_STYLE_UDPHDR
#define UDP_SOURCE(h) (h)->uh_sport
#define UDP_DEST(h) (h)->uh_dport
#define UDP_LEN(h) (h)->uh_ulen
#define UDP_CHECK(h) (h)->uh_sum
#else
#define UDP_SOURCE(h) (h)->source
#define UDP_DEST(h) (h)->dest
#define UDP_LEN(h) (h)->len
#define UDP_CHECK(h) (h)->check
#endif

namespace dns_analyzer {

    using std::dec;
    using std::invalid_argument;
    using std::istringstream;

    class abstract_accessor_type {
    public:
        virtual value_ref operator()(network_packet const &packet) const = 0;
    };

    template <class T>
    class accessor_type : public abstract_accessor_type {
    public:
        typedef value_ref (T::*field_accessor_type)() const;

        accessor_type(field_accessor_type field_accessor)
            : _field_accessor(field_accessor)
            {}

        virtual value_ref operator()(network_packet const &packet) const {
            T const *p = find_down<T>(&packet);
            return p ? (p->*_field_accessor)() : null_value;
        }
    private:
        field_accessor_type _field_accessor;
    };

    struct name_protocol_field_type {
        string name;
        abstract_accessor_type *accessor;
    };

    static name_protocol_field_type name_protocol_field_table[] = {
        { "ip.src", new accessor_type<ip_packet>(&ip_packet::field_src) },
        { "ip.dst", new accessor_type<ip_packet>(&ip_packet::field_dst) },
        { "udp.src", new accessor_type<udp_packet>(&udp_packet::field_src) },
        { "udp.dst", new accessor_type<udp_packet>(&udp_packet::field_dst) },
        { "udp.src-port", new accessor_type<udp_packet>(&udp_packet::field_src_port) },
        { "udp.dst-port", new accessor_type<udp_packet>(&udp_packet::field_dst_port) },
        { "dns.qcount", new accessor_type<dns_packet>(&dns_packet::field_qcount) },
        { "dns.type", new accessor_type<dns_packet>(&dns_packet::field_type) },
        { "dns.class", new accessor_type<dns_packet>(&dns_packet::field_class) },
        { "dns.domain", new accessor_type<dns_packet>(&dns_packet::field_domain) },
        { "dns.query", new accessor_type<dns_packet>(&dns_packet::field_query) },
        { "dns.authoritative", new accessor_type<dns_packet>(&dns_packet::field_authoritative) },
        { "dns.truncated", new accessor_type<dns_packet>(&dns_packet::field_truncated) },
        { "dns.rcode", new accessor_type<dns_packet>(&dns_packet::field_rcode) },
        { "dns.server", new accessor_type<dns_packet>(&dns_packet::field_server) },
        { "dns.client", new accessor_type<dns_packet>(&dns_packet::field_client) },
    };

    packet_field_type::packet_field_type(string const &name)
        : _name(name), _accessor(0)
    {
        size_t table_size = (sizeof(name_protocol_field_table)
                             / sizeof(name_protocol_field_table[0]));
        for (size_t i = 0; i < table_size; ++i) {
            if (_name == name_protocol_field_table[i].name) {
                _accessor = name_protocol_field_table[i].accessor;
                break;
            }
        }

        if (!_accessor) {
            throw invalid_argument(_name);
        }
    }

    packet_field_type::packet_field_type(packet_field_type const &rhs)
        : _name(rhs._name), _accessor(rhs._accessor)
    {
    }

    packet_field_type const &
    packet_field_type::operator=(packet_field_type const &rhs) {
        if (this != &rhs) {
            _name = rhs._name;
            _accessor = rhs._accessor;
        }
        return *this;
    }

    string
    packet_field_type::to_string() const {
        return _name;
    }

    value_ref
    packet_field_type::get(network_packet const &packet) const
    {
        return (*_accessor)(packet);
    }

    packet_field_list_type
    make_field_list(string const &fields) {
        packet_field_list_type result;
        istringstream in(fields);
        while (in) {
            string name;
            getline(in, name, ',');
            if (in) {
                result.push_back(packet_field_type(name));
            }
        }
        return result;
    }

    packet_data::packet_data(
        u_char const *data,
        size_t captured_size,
        size_t original_size,
        timeval const &timestamp)
        :
        _managed(false),
        _header(),
        _data(data)
    {
        _header.caplen = captured_size;
        _header.len = original_size;
        _header.ts = timestamp;
        if (_data == 0) {
            throw std::logic_error("null data");
        }
        if (captured_size > original_size) {
            throw std::logic_error("captured_size > original_size");
        }
    }

    packet_data::packet_data(
        u_char const *data,
        pcap_pkthdr const &header)
        : _managed(false),
          _header(header),
          _data(data)
    {
        if (_data == 0) {
            throw std::logic_error("null data");
        }
        if (captured_size() > original_size()) {
            throw std::logic_error("captured_size > original_size");
        }
    }

    packet_data::packet_data(packet_data const &rhs)
        : _managed(rhs._managed),
          _header(rhs._header),
          _data(rhs._data)
    {
        if (_managed) {
            u_char *data = new u_char[rhs.captured_size()];
            std::copy(rhs.begin(), rhs.end(), data);
            _data = data;
        }
    }

    packet_data::~packet_data() {
        if (_managed) {
            delete[] _data;
        }
    }

    packet_data const &
    packet_data::operator=(packet_data const &rhs) {
        if (this == &rhs) {
            return *this;
        }

        if (_managed) {
            delete[] _data;
        }

        _managed = rhs._managed;
        _header = rhs._header;
        _data = rhs._data;

        if (_managed) {
            u_char *data = new u_char[rhs.captured_size()];
            std::copy(rhs.begin(), rhs.end(), data);
            _data = data;
        }

        return *this;
    }

    bool
    packet_data::operator==(packet_data const &rhs) const {
        if (this == &rhs) {
            return true;
        }

        if (_header.caplen != rhs._header.caplen
            || _header.len != rhs._header.len
            || _header.ts.tv_sec != rhs._header.ts.tv_sec
            || _header.ts.tv_usec != rhs._header.ts.tv_usec)
        {
            return false;
        }

        return memcmp(_data, rhs._data, captured_size()) == 0;
    }

    bool
    packet_data::operator!=(packet_data const &rhs) const {
        return !this->operator==(rhs);
    }

    packet_data
    packet_data::slice(size_t first, size_t last) const {
        if (first > captured_size() || last > captured_size()) {
            throw std::out_of_range("packet_data::slice");
        }
        return packet_data(begin() + first,
                           last - first,
                           original_size() - first,
                           timestamp());
    }

    u_char const &
    packet_data::at(size_t index) const {
        if (index >= captured_size()) {
            throw std::out_of_range("packet_data::at");
        }
        return _data[index];
    }

    u_char const &
    packet_data::operator[](size_t index) const {
        return _data[index];
    }

    double
    packet_data::timestamp_as_double() const {
        return double(timestamp().tv_sec) + double(timestamp().tv_usec) / 1.0e6;
    }

    void
    packet_data::make_managed() {
        if (_managed) {
            return;
        }

        _managed = true;
        u_char *data = new u_char[captured_size()];
        std::copy(begin(), end(), data);
        _data = data;
    }

    network_packet::network_packet(packet_data const &data,
                                   network_packet const *container)
        : _data(data), _container(container)
    {
    }

    network_packet::~network_packet() {
    }

    network_packet *
    network_packet::deep_copy() const {
        network_packet *result = shallow_copy();
        result->_data.make_managed();
        return result;
    }

    network_packet const *
    network_packet::outermost_container() const {
        network_packet const *result = this;
        while (result->container()) {
            result = result->container();
        }
        return result;
    }

    network_packet const *
    network_packet::innermost_payload() const {
        network_packet const *result = this;
        while (result->payload()) {
            result = result->payload();
        }
        return result;
    }

    void
    network_packet::print_info(ostream &out) const {
        size_t indent = 0;
        out << dec << data().captured_size() << "/"
            << data().original_size() << " ";
        for (network_packet const *p = this; p != 0; p = p->payload()) {
            ++indent;
            p->print_payload_info(out, indent);
            if (p->payload()) {
                out << ":\n" << string(indent, ' ');
            }
        }
        out << "\n";
    }

    static inline struct ether_header const *
    ethernet_packet_header(ethernet_packet const *packet) {
        return reinterpret_cast<struct ether_header const *>(
            packet->data().begin());
    }

    ethernet_packet::ethernet_packet(packet_data const &data,
                                     network_packet const *container)
        : network_packet(data, container), _payload(0)
    {
    }

    ethernet_packet::~ethernet_packet() {
        delete _payload;
    }

    void
    ethernet_packet::print_payload_info(ostream &out, size_t /*indent*/) const {
        out << "eth";
    }

    ethernet_packet *
    ethernet_packet::shallow_copy() const {
        if (container() != 0) {
            throw std::logic_error("cannot copy payload packet");
        }
        return new ethernet_packet(data());
    }

    bool
    ethernet_packet::good() const {
        if (data().captured_size() < ETHERMIN) {
            return false;
        }

        u_short t = type();
        return t == ETHERNET_TYPE_IPV4 || t == ETHERNET_TYPE_IPV6;
    }

    size_t
    ethernet_packet::header_size() const {
        return sizeof(struct ether_header);
    }

    network_packet const *
    ethernet_packet::payload() const {
        if (_payload) {
            return _payload;
        }

        if (!good()) {
            return 0;
        }

        switch (type()) {
        case ETHERNET_TYPE_IPV4:
            _payload = new ipv4_packet(body(), this);
            break;
        case ETHERNET_TYPE_IPV6:
            _payload = new ipv6_packet(body(), this);
            break;
        default:
            throw unknown_packet_type("unrecognized ethernet payload");
        }

        return _payload;
    }

    u_short
    ethernet_packet::type() const {
        return ntohs(ethernet_packet_header(this)->ether_type);
    }

    internet_address::internet_address(sa_family_t af,
                                       string const &text,
                                       in_port_t port)
    {
        int rc;
        switch (af) {
        case AF_INET:
#if defined(HAVE_SOCKADDR_IN_SIN_LEN)
            _addr_in.sin_len = sizeof(_addr_in);
#endif
            _addr_in.sin_family = af;
            _addr_in.sin_port = port;
            rc = inet_pton(af, text.c_str(), &_addr_in.sin_addr);
            std::fill_n(_addr_in.sin_zero, sizeof(_addr_in.sin_zero), '\0');
            break;
        case AF_INET6:
#if defined(HAVE_SOCKADDR_IN6_SIN6_LEN)
            _addr_in6.sin6_len = sizeof(_addr_in6);
#endif
            _addr_in6.sin6_family = af;
            _addr_in6.sin6_port = port;
            _addr_in6.sin6_flowinfo = 0;
            rc = inet_pton(af, text.c_str(), &_addr_in6.sin6_addr);
            _addr_in6.sin6_scope_id = 0;
            break;
        default:
            throw std::logic_error("internet_address: address family");
        }

        if (rc == 0) {
            throw std::domain_error("internet_address: address unparseable");
        } else if (rc == -1) {
            throw std::runtime_error("internet_address: inet_pton failed");
        }
    }

    internet_address::internet_address(in_addr addr, in_port_t port) {
#if defined(HAVE_SOCKADDR_IN_SIN_LEN)
        _addr_in.sin_len = sizeof(_addr_in);
#endif
        _addr_in.sin_family = AF_INET;
        _addr_in.sin_port = port;
        _addr_in.sin_addr = addr;
        std::fill_n(_addr_in.sin_zero, sizeof(_addr_in.sin_zero), '\0');
    }

    internet_address::internet_address(in6_addr const &addr, in_port_t port) {
#if defined(HAVE_SOCKADDR_IN6_SIN6_LEN)
        _addr_in6.sin6_len = sizeof(_addr_in6);
#endif
        _addr_in6.sin6_family = AF_INET6;
        _addr_in6.sin6_port = port;
        _addr_in6.sin6_addr = addr;
    }

    internet_address::internet_address(sockaddr const &rhs) {
        switch (_addr.sa_family != AF_INET) {
        case AF_INET:
            _addr_in = (sockaddr_in const &) rhs;
            break;
        case AF_INET6:
            _addr_in6 = (sockaddr_in6 const &) rhs;
            break;
        default:
            throw std::logic_error("internet_address: bad address family");
        }
    }

    internet_address::internet_address(sockaddr_in const &rhs)
        : _addr_in(rhs)
    {
        if (_addr.sa_family != AF_INET) {
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    internet_address::internet_address(sockaddr_in6 const &rhs)
        : _addr_in6(rhs)
    {
        if (_addr.sa_family != AF_INET6) {
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    internet_address::internet_address(internet_address const &rhs) {
        switch (rhs.address_family()) {
        case AF_INET:
            std::memcpy(&_addr_in, &rhs._addr_in, sizeof(_addr_in));
            break;
        case AF_INET6:
            std::memcpy(&_addr_in6, &rhs._addr_in6, sizeof(_addr_in6));
            break;
        default:
            throw std::logic_error("internet_address: bad address family");
        }
    }

    internet_address const &
    internet_address::operator=(internet_address const &rhs) {
        if (this != &rhs) {
            switch (rhs._addr.sa_family) {
            case AF_INET:
                _addr_in = rhs._addr_in;
                break;
            case AF_INET6:
                _addr_in6 = rhs._addr_in6;
                break;
            default:
                throw std::runtime_error("internet_address: bad address family");
            }
        }

        return *this;
    }

    bool
    internet_address::operator==(internet_address const &rhs) const {
        if (this == &rhs) {
            return true;
        }

        if (address_family() != rhs.address_family()) {
            return false;
        }

        switch (address_family()) {
        case AF_INET:
            return std::memcmp(&_addr_in, &rhs._addr_in, sizeof(_addr_in)) == 0;
        case AF_INET6:
            return (IN6_ARE_ADDR_EQUAL(&_addr_in6.sin6_addr,
                                       &rhs._addr_in6.sin6_addr)
                    && _addr_in6.sin6_port == rhs._addr_in6.sin6_port);
        default:
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    bool
    internet_address::operator!=(internet_address const &rhs) const {
        return !(*this == rhs);
    }

    bool
    internet_address::operator<(internet_address const &rhs) const {
        if (this == &rhs) {
            return false;
        }

        if (address_family() != rhs.address_family()) {
            return address_family() < rhs.address_family();
        }

        switch (address_family()) {
        case AF_INET:
            return std::memcmp(&_addr_in, &rhs._addr_in, sizeof(_addr_in)) < 0;
        case AF_INET6: {
            int cmp = std::memcmp(&_addr_in6.sin6_addr,
                                  &rhs._addr_in6.sin6_addr,
                                  sizeof(_addr_in6.sin6_addr));
            return (cmp < 0
                    ? true
                    : _addr_in6.sin6_port < rhs._addr_in6.sin6_port);
        }
        default:
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    bool
    internet_address::operator<=(internet_address const &rhs) const {
        return this->operator<(rhs) || this->operator==(rhs);
    }

    bool
    internet_address::operator>(internet_address const &rhs) const {
        return !this->operator<=(rhs);
    }

    bool
    internet_address::operator>=(internet_address const &rhs) const {
        return !this->operator<(rhs);
    }

    sa_family_t
    internet_address::address_family() const {
        return _addr.sa_family;
    }

    sockaddr_in const &
    internet_address::as_in4() const {
        if (_addr.sa_family != AF_INET) {
            throw std::logic_error("internet_address: bad address family");
        }
        return _addr_in;
    }

    sockaddr_in6 const &
    internet_address::as_in6() const {
        if (_addr.sa_family != AF_INET6) {
            throw std::logic_error("internet_address: bad address family");
        }
        return _addr_in6;
    }

    void
    internet_address::port(in_port_t port) {
        switch (_addr.sa_family) {
        case AF_INET:
            _addr_in.sin_port = port;
            break;
        case AF_INET6:
            _addr_in6.sin6_port = port;
            break;
        default:
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    in_port_t
    internet_address::port() const {
        switch (_addr.sa_family) {
        case AF_INET:
            return _addr_in.sin_port;
        case AF_INET6:
            return _addr_in6.sin6_port;
        default:
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    string
    internet_address::to_string() const {
        char tmp[100];
        char const *rc;
        switch (_addr.sa_family) {
        case AF_INET:
            rc = inet_ntop(_addr_in.sin_family, &_addr_in.sin_addr,
                           tmp, sizeof(tmp));
            break;
        case AF_INET6:
            rc = inet_ntop(_addr_in6.sin6_family, &_addr_in6.sin6_addr,
                           tmp, sizeof(tmp));
            break;
        default:
            throw std::runtime_error("internet_address: bad address family");
        }

        if (rc == 0) {
            throw std::runtime_error("internet_address: inet_ntop failed");
        }

        return string(rc);
    }

    size_t
    internet_address::hash_func() const {
        switch (_addr.sa_family) {
        case AF_INET:
            return ((size_t) _addr_in.sin_addr.s_addr) ^ ((size_t) _addr_in.sin_port);
        case AF_INET6: {
            size_t result = (size_t) _addr_in6.sin6_port;
            for (size_t i = 0; i < sizeof(_addr_in6.sin6_addr.s6_addr); ++i)
                result = result * 5 + (size_t) _addr_in6.sin6_addr.s6_addr[i];
            return result;
        }
        default:
            throw std::runtime_error("internet_address: bad address family");
        }
    }

    ostream &
    operator<<(ostream &out, internet_address const &rhs) {
        return out << rhs.to_string();
    }

    ip_packet::ip_packet(packet_data const &data,
                         network_packet const *container)
        : network_packet(data, container)
    {
    }

    ip_packet::~ip_packet() {
    }

    value_ref
    ip_packet::field_src() const {
        return typed_value<internet_address>(source_address());
    }

    value_ref
    ip_packet::field_dst() const {
        return typed_value<internet_address>(dest_address());
    }

    static inline struct ip const *
    ipv4_packet_header(ipv4_packet const *packet) {
        return reinterpret_cast<struct ip const *>(packet->data().begin());
    }

    ipv4_packet::ipv4_packet(packet_data const &data,
                             network_packet const *container)
        : ip_packet(data, container), _payload(0)
    {
    }

    ipv4_packet::~ipv4_packet() {
        delete _payload;
    }

    void
    ipv4_packet::print_payload_info(ostream &out, size_t /*indent*/) const {
        out << "ipv4 " << source_address() << " -> " << dest_address();
    }

    ipv4_packet *
    ipv4_packet::shallow_copy() const {
        if (container() != 0) {
            throw std::logic_error("cannot copy payload packet");
        }
        return new ipv4_packet(data());
    }

    bool
    ipv4_packet::good() const {
        if (data().captured_size() <= sizeof(struct ip)) {
            return false;
        }

        struct ip const *h = ipv4_packet_header(this);
        return h->ip_v == IPVERSION && h->ip_hl <= data().captured_size();
    }

    size_t
    ipv4_packet::header_size() const {
        assert(good());
        return ipv4_packet_header(this)->ip_hl * 4;
    }

    network_packet const *
    ipv4_packet::payload() const {
        if (_payload) {
            return _payload;
        }

        if (!good()) {
            return false;
        }

        switch (ipv4_packet_header(this)->ip_p) {
        case IPPROTO_UDP:
            _payload = new udp_packet(body(), this);
            break;
        default:
            break;
        }
        return _payload;
    }

    internet_address
    ipv4_packet::source_address(in_port_t port) const {
        return internet_address(ipv4_packet_header(this)->ip_src, port);
    }

    internet_address
    ipv4_packet::dest_address(in_port_t port) const {
        return internet_address(ipv4_packet_header(this)->ip_dst, port);
    }

    static inline struct ip6_hdr const *
    ipv6_packet_header(ipv6_packet const *packet) {
        return reinterpret_cast<struct ip6_hdr const *>(
            packet->data().begin());
    }

    ipv6_packet::ipv6_packet(packet_data const &data,
                             network_packet const *container)
        : ip_packet(data, container), _payload(0)
    {
    }

    ipv6_packet::~ipv6_packet() {
        delete _payload;
    }

    void
    ipv6_packet::print_payload_info(ostream &out, size_t /*indent*/) const {
        out << "ipv6 " << source_address() << " -> " << dest_address();
    }

    ipv6_packet *
    ipv6_packet::shallow_copy() const {
        if (container() != 0) {
            throw std::logic_error("cannot copy payload packet");
        }
        return new ipv6_packet(data());
    }

    bool
    ipv6_packet::good() const {
        if (data().captured_size() <= sizeof(struct ip6_hdr)) {
            return false;
        }

        struct ip6_hdr const *h = ipv6_packet_header(this);
        return (h->ip6_vfc & IPV6_VERSION_MASK) == IPV6_VERSION;
    }

    size_t
    ipv6_packet::header_size() const {
        //TODO: deal with extension headers.
        assert(good());
        return sizeof(struct ip6_hdr);
    }

    network_packet const *
    ipv6_packet::payload() const {
        if (_payload) {
            return _payload;
        }

        if (!good()) {
            return false;
        }

        switch (ipv6_packet_header(this)->ip6_nxt) {
        case IPPROTO_UDP:
            _payload = new udp_packet(body(), this);
            break;
        default:
            break;
        }
        return _payload;
    }

    internet_address
    ipv6_packet::source_address(in_port_t port) const {
        return internet_address(ipv6_packet_header(this)->ip6_src, port);
    }

    internet_address
    ipv6_packet::dest_address(in_port_t port) const {
        return internet_address(ipv6_packet_header(this)->ip6_dst, port);
    }

    struct udphdr const *
    udp_packet_header(udp_packet const *packet) {
        return reinterpret_cast<struct udphdr const *>(packet->data().begin());
    }

    udp_packet::udp_packet(packet_data const &data, ip_packet const *container)
        : network_packet(data, container), _payload(0)
    {
    }

    udp_packet::~udp_packet() {
        delete _payload;
    }

    void
    udp_packet::print_payload_info(ostream &out, size_t /*indent*/) const {
        out << "udp " << dec << source_port() << " -> " << dest_port();
    }

    udp_packet *
    udp_packet::shallow_copy() const {
        throw std::logic_error("cannot copy payload packet");
    }

    bool
    udp_packet::good() const {
        if (data().captured_size() < sizeof(struct udphdr)) {
            return false;
        }

        return (ntohs(UDP_LEN(udp_packet_header(this)))
                >= sizeof(struct udphdr));
    }

    size_t
    udp_packet::header_size() const {
        return sizeof(struct udphdr);
    }

    network_packet const *
    udp_packet::payload() const {
        if (_payload) {
            return _payload;
        }

        if (!good()) {
            return 0;
        }

        in_port_t sport = source_port();
        in_port_t dport = dest_port();
        if (sport == NS_DEFAULTPORT || dport == NS_DEFAULTPORT) {
            _payload = new dns_packet(body(), this);
        } else {
            throw unknown_packet_type("unrecognized UDP payload type");
        }

        return _payload;
    }

    internet_address
    udp_packet::source_address() const {
        ip_packet const *ip = find_container<ip_packet>(this);
        assert(ip != 0);
        return ip->source_address(source_port());
    }

    internet_address
    udp_packet::dest_address() const {
        ip_packet const *ip = find_container<ip_packet>(this);
        assert(ip != 0);
        return ip->dest_address(dest_port());
    }

    in_port_t
    udp_packet::source_port() const {
        return ntohs(UDP_SOURCE(udp_packet_header(this)));
    }

    in_port_t
    udp_packet::dest_port() const {
        return ntohs(UDP_DEST(udp_packet_header(this)));
    }

    value_ref
    udp_packet::field_src() const {
        return typed_value<internet_address>(source_address());
    }

    value_ref
    udp_packet::field_dst() const {
        return typed_value<internet_address>(dest_address());
    }

    value_ref
    udp_packet::field_src_port() const {
        return typed_value<in_port_t>(source_port());
    }

    value_ref
    udp_packet::field_dst_port() const {
        return typed_value<in_port_t>(dest_port());
    }
}
