// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef PACKET_HH
#define PACKET_HH

#include "fast_map.hh"

#include <cassert>
#include <map>
#include <iosfwd>
#include <stdexcept>
#include <string>
#include <typeinfo>
#include <vector>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
extern "C" {
#include <pcap.h>
}

namespace dns_analyzer {

    using std::ostream;
    using std::string;
    using std::vector;

    typedef vector<string> field_name_type;
    typedef vector<field_name_type> field_name_list_type;

    class value_ref;

    class network_packet;
    class ethernet_packet;
    class ipv4_packet;
    class ipv6_packet;
    class dns_packet;

    class abstract_accessor_type;

    /**
     * The packet_field_type class is used to specify both the packet
     * protocol and field name of a particular field in a network
     * packet.  This is used to look up a field at run-time.
     */
    class packet_field_type {
    public:
        /**
         * Construct a new packet field.
         *
         * @param name the name of the packet field.
         *
         * @throws std::invalid_argument if the name is not valid.
         * The message contains the invalid field name.
         */
        explicit packet_field_type(string const &name);

        /**
         * Copy constructor.
         */
        packet_field_type(packet_field_type const &rhs);

        /**
         * Assignment.
         */
        packet_field_type const &operator=(packet_field_type const &rhs);

        /**
         * Convert this instance to a human readable string.
         *
         * @return A string of the form: "protocol.field".
         */
        string to_string() const;

        /**
         * Read this field from the indicated packet and return the
         * result as a generic value.
         *
         * @param packet the packet to read the field from.
         *
         * @return The field value as a reference to a generic value.
         */
        value_ref get(network_packet const &packet) const;

    private:
        string _name;
        abstract_accessor_type const *_accessor;
    };

    typedef vector<packet_field_type> packet_field_list_type;

    /**
     * Parse a comma separated list of field names into a
     * packet_field_list_type.
     *
     * @throws std::invalid_argument if the name is not valid.
     * The message contains the invalid field name.
     */
    packet_field_list_type make_field_list(string const &fields);

    /**
     * Exception raised when an unknown network packet type is
     * found.
     */
    class unknown_packet_type : public std::runtime_error {
    public:
        explicit unknown_packet_type(string const &message)
            : std::runtime_error(message) {}
    };

    /**
     * Represents the raw data of a network packet.  The main
     * responsibility of this class is to correctly deallocate the
     * data if the data was dynamically allocated.
     */
    class packet_data {
    public:
        /**
         * A random access iterator for the packet's data.
         */
        typedef u_char const *const_iterator;

        /**
         * Construct a new packet data based on the provided
         * information.
         *
         * @param data the data [object, or until the make_managed is
         * invoked on this instance].
         *
         * @param captured_size the size of the data in octets.
         *
         * @param original_size the original size of the packet.
         *
         * @param timestamp the time the packet was captured.
         */
        packet_data(u_char const *data, size_t captured_size,
                    size_t original_size, timeval const &timestamp);

        /**
         * Construct a new packet data based on the provided
         * information.
         *
         * @param data the data [object, or until the make_managed is
         * invoked on this instance].
         *
         * @param header the packet header as returned by the pcap
         * library [function].
         */
        packet_data(u_char const *data, pcap_pkthdr const &header);

        /**
         * Copy constructor.  If the rhs is managed the new instance
         * will also be managed and have its own, private copy of the
         * data.
         *
         * @param rhs the original packet.
         */
        packet_data(packet_data const &rhs);

        /**
         * Destructor.  The destructor deallocates the data if the
         * packet data is managed by this instance.
         */
        ~packet_data();

        /**
         * Assignment.  If this instance's data is managed it is
         * deallocated before acquiring the data from the original.
         * If the original instance's data is managed this instance
         * will have its own, private copy of the original's data.
         */
        packet_data const &operator=(packet_data const &rhs);

        /**
         * Test for equality.  Two packets are equal if they have the
         * same data, captured size, original size, and timestamp.
         */
        bool operator==(packet_data const &rhs) const;

        /**
         * Test for inequality.  See operator==().
         */
        bool operator!=(packet_data const &rhs) const;

        /**
         * Create an unmanaged packet_data representing the data
         * [first, last).
         *
         * @param first the position of the first octet.  The octet in
         * this position is included in the result.
         *
         * @param last the position of the last octet.  The octet in
         * this position is not included in the result.
         *
         * @throws std::out_of_range if first or last is not in the
         * range [ 0, captured_size() ].
         */
        packet_data slice(size_t first, size_t last) const;

        /**
         * Access the octet at specified position.
         *
         * @param index the position of the octet.
         *
         * @throws std::out_of_range if index is not in the range [ 0,
         * captured_size() ).
         */
        u_char const &at(size_t index) const;

        /**
         * Access the octet at the specified position.  The result is
         * undefined if the position is outside of the range [ 0,
         * captured_size() ).
         *
         * @param index the position of the octet.
         */
        u_char const &operator[](size_t index) const;

        /**
         * The number of octets present in this packet data.
         */
        size_t captured_size() const { return _header.caplen; }

        /**
         * The number of octets in the original packet.
         */
        size_t original_size() const { return _header.len; }

        /**
         * The time this packet was captured.
         */
        timeval const &timestamp() const { return _header.ts; }

        /**
         * The time this packet was captured.
         */
        double timestamp_as_double() const;

        /**
         * The pcap packet header.
         */
        pcap_pkthdr const &header() const { return _header; }

        /**
         * An iterator pointing to the first octet in the packet data.
         */
        const_iterator begin() const { return _data; }

        /**
         * An iterator pointing to one octet past the last octet in
         * the packet data.
         */
        const_iterator end() const { return _data + captured_size(); }

        /**
         * Is the captured_size() smaller than the original_size() ?
         */
        bool is_truncated() const {
            return captured_size() < original_size();
        }

        /**
         * Does this instance keep its own, private copy of the actual
         * data?
         */
        bool is_managed() const { return _managed; }

        /**
         * Ensure this instance keeps its own, private copy of the
         * actual data.  If this instance is not already managed a
         * copy of the shared data is made and this instance is now
         * considered to be managed.
         */
        void make_managed();
    private:
        bool _managed;
        pcap_pkthdr _header;
        u_char const *_data;
    };

    /**
     * An abstract class that serves as the ancestor to specific kinds
     * of network packets.  This class wraps a packet_data instance
     * and provides higher-level, protocol specific access to the raw
     * packet data.
     */
    class network_packet {
    public:
        /**
         * Construct a new network packet.
         *
         * @param data the packet's data [object or until the data is
         * managed].
         *
         * @param container this packet's containing packet (if any)
         * [object].
         */
        network_packet(packet_data const &data,
                       network_packet const *container = 0);

        /**
         * Virtual destructor.
         */
        virtual ~network_packet();

        /**
         * The captured data of the network packet.
         */
        packet_data const &data() const { return _data; }

        /**
         * The body of the network packet, which usually starts right
         * after the header.  Note that iff payload() != 0 then
         * this->body() == this->payload()->data().
         */
        packet_data body() const {
            return _data.slice(header_size(), _data.captured_size());
        }

        /**
         * The packet that contains this packet, if any.
         */
        network_packet const *container() const { return _container; }

        /**
         * Print the packet in a human readable format on the
         * indicated output stream.
         *
         * @param out the output stream to print this instance on
         * [function].
         */
        void print_info(ostream &out) const;

        /**
         * Print the payload information of this packet in a human
         * readable format on the indicated output stream.
         *
         * @param out the output stream to print the payload
         * information on [function].
         *
         * @param indent the indentation level to start on.
         */
        virtual void print_payload_info(ostream &out, size_t indent = 0) const
            = 0;

        /**
         * Create a new, shallow copy of this network packet.  The
         * caller is responsible for deleting the copy.
         */
        virtual network_packet *shallow_copy() const = 0;

        /**
         * Create a new copy of this network packet with its own,
         * private copy of the packet's data.
         */
        network_packet *deep_copy() const;

        /**
         * True if this packet is known to be valid.
         */
        virtual bool good() const = 0;

        /**
         * The size of the header of a packet.  Some protocols do not
         * have a separation between headers and payload and in this
         * case 0 should be returned.
         */
        virtual size_t header_size() const = 0;

        /**
         * The packet representing the payload of this network packet.
         * Not all network protocols support payloads (such as DNS)
         * and in this case 0 is returned.  Note that iff payload() !=
         * 0 then this->payload()->container() == this.
         */
        virtual network_packet const *payload() const = 0;

        /**
         * Find the outermost container.  Returns 'this' if this packet
         * is not contained in another packet (in other words, 'this'
         * is the outermost container).
         */
        network_packet const *outermost_container() const;

        /**
         * Find the innermost payload.  Returns 'this' if this
         * packet does not have a payload (in other words, 'this' is
         * the innermost payload).
         */
        network_packet const *innermost_payload() const;

    private:
        packet_data _data;
        network_packet const *_container;

        /*
         * Prohibit copying.
         */
        network_packet(network_packet const &rhs);

        /*
         * Prohibit assignment.
         */
        network_packet const &operator=(network_packet const &rhs);
    };

    /**
     * Like <code>find_container</code> but starts at
     * <code>packet</code> before looking at containers.
     */
    template <class T> T const *find_up(network_packet const *packet) {
        try {
            for (network_packet const *c = packet; c != 0; c = c->container()) {
                if (T const *p = dynamic_cast<T const *>(c)) {
                    return p;
                }
            }
        } catch (unknown_packet_type &ex) {
        }
        return 0;
    }

    /**
     * Recurses through all the containers to find the first
     * container of type T using dynamic cast.  0 is returned if
     * no appropriate packet contains <code>packet</code>.
     */
    template <class T> T const *find_container(network_packet const *packet) {
        assert(packet != 0);
        return find_up<T>(packet->container());
    }

    /**
     * Like <code>find_payload</code> but starts at
     * <code>packet</code> before looking at payloads.
     */
    template <class T> T const *find_down(network_packet const *packet) {
        try {
            for (network_packet const *c = packet; c != 0; c = c->payload()) {
                if (T const *p = dynamic_cast<T const *>(c)) {
                    return p;
                }
            }
        } catch (unknown_packet_type &ex) {
        }
        return 0;
    }

    /**
     * Recurses through all the payloads to find the first payload
     * of type T using dynamic cast.  0 is returned if no
     * appropriate packet is contained in <code>packet</code>.
     */
    template <class T> T const *find_payload(network_packet const *packet) {
        assert(packet != 0);
        return find_down<T>(packet->payload());
    }

    /**
     * Ethernet packet.
     */
    class ethernet_packet : public network_packet {
    public:
        /**
         * Construct a new ethernet packet.
         *
         * @param data the data of the ethernet packet.
         *
         * @param container the containing packet (if any).
         */
        explicit ethernet_packet(packet_data const &data,
                                 network_packet const *container = 0);

        /**
         * Virtual destructor.
         */
        virtual ~ethernet_packet();

        virtual void print_payload_info(ostream &out, size_t indent = 0) const;

        virtual ethernet_packet *shallow_copy() const;
        virtual bool good() const;
        virtual size_t header_size() const;
        virtual network_packet const *payload() const;

        /**
         * The type of the ethernet payload in host byte-order.
         */
        u_short type() const;

    private:
        mutable network_packet const *_payload;
    };

    /**
     * This class represents an internet address with port number.
     * The internet address can be either an IPv4 or IPv6 address.
     */
    class internet_address {
    public:
        /**
         * Construct a new internet address from a string
         * representation.
         *
         * @param af the address family (AF_INET or AF_INET6) of the
         * internet address.
         *
         * @param text the string representation of the IP address.
         *
         * @param port the optional port number of the IP address.
         */
        internet_address(sa_family_t af, string const &text,
                         in_port_t port = 0);

        /**
         * Construct a new IPv4 address.
         *
         * @param addr the IPv4 address.
         *
         * @param port the optional port number.
         */
        explicit internet_address(in_addr addr, in_port_t port = 0);

        /**
         * Construct a new IPv6 address.
         *
         * @param addr the IPv6 address.
         *
         * @param port the optional port number.
         */
        explicit internet_address(in6_addr const &addr, in_port_t port = 0);

        /**
         * Construct a new IP address.
         *
         * @param rhs the IP address.  The sa_family member determines
         * whether this is an IPv4 or IPv6 address.
         */
        explicit internet_address(sockaddr const &rhs);

        /**
         * Construct a new IPv4 address.
         *
         * @param rhs the IPv4 address.
         */
        explicit internet_address(sockaddr_in const &rhs);

        /**
         * Construct a new IPv6 address.
         *
         * @param rhs the IPv6 address.
         */
        explicit internet_address(sockaddr_in6 const &rhs);

        /**
         * Copy constructor.
         */
        internet_address(internet_address const &rhs);

        /**
         * Assignment.
         */
        internet_address const &operator=(internet_address const &rhs);

        /**
         * Equality.
         */
        bool operator==(internet_address const &rhs) const;

        /**
         * Inequality.
         */
        bool operator!=(internet_address const &rhs) const;

        /**
         * IPv4 address are considered to be less than IPv6 addresses.
         * IP addresses are compared numerically.
         */
        bool operator<(internet_address const &rhs) const;

        /**
         * See operator<().
         */
        bool operator<=(internet_address const &rhs) const;

        /**
         * See operator<().
         */
        bool operator>(internet_address const &rhs) const;

        /**
         * See operator<().
         */
        bool operator>=(internet_address const &rhs) const;

        /**
         * The address family (AF_INET or AF_INET6).
         */
        sa_family_t address_family() const;

        /**
         * Convert the internet address to an IPv4 sockaddr_in
         * representation.  The internet address must be an IPv4
         * address.
         */
        sockaddr_in const &as_in4() const;

        /**
         * Convert the internet address to an IPv6 sockaddr_in6
         * representation.  The internet address must be an IPv6
         * address.
         */
        sockaddr_in6 const &as_in6() const;

        /**
         * Set the port number.
         */
        void port(in_port_t port);

        /**
         * The port number.
         */
        in_port_t port() const;

        /**
         * The string representation of the current internet address
         * without the port number.
         */
        string to_string() const;

        /**
         * Calculate a hash function.
         */
        size_t hash_func() const;
    private:
        union {
            sockaddr _addr;
            sockaddr_in _addr_in;
            sockaddr_in6 _addr_in6;
        };
    };

    /**
     * Put the string representation of an internet_address on the
     * specified ostream.
     */
    ostream &operator<<(ostream &out, internet_address const &rhs);

    /**
     * Abstract class representing Internet Protocol (IP) packets.
     * Subclasses are defined for both IPv4 and IPv6.
     */
    class ip_packet : public network_packet {
    public:
        /**
         * Constructor.
         *
         * @param data the data of the IP packet.
         *
         * @param container the containing packet (if any).
         */
        ip_packet(packet_data const &data, network_packet const *container = 0);

        /**
         * Virtual destructor.
         */
        virtual ~ip_packet();

        /**
         * The source address.
         */
        virtual internet_address source_address(in_port_t port = 0) const = 0;

        /**
         * The destination address.
         */
        virtual internet_address dest_address(in_port_t port = 0) const = 0;

        /**
         * Access the source address field as a value_ref.
         */
        value_ref field_src() const;

        /**
         * Access the destination address field as a value_ref.
         */
        value_ref field_dst() const;
    };

    /**
     * This class represents IPv4 packets.
     */
    class ipv4_packet : public ip_packet {
    public:
        /**
         * Constructor.
         *
         * @param data the data of the IPv4 packet.
         *
         * @param container the containing packet (if any).
         */
        ipv4_packet(packet_data const &data,
                    network_packet const *container = 0);

        /**
         * Virtual destructor.
         */
        virtual ~ipv4_packet();

        virtual void print_payload_info(ostream &out, size_t indent = 0) const;

        virtual ipv4_packet *shallow_copy() const;

        virtual bool good() const;
        virtual size_t header_size() const;
        virtual network_packet const *payload() const;

        virtual internet_address source_address(in_port_t port = 0) const;
        virtual internet_address dest_address(in_port_t port = 0) const;
    private:
        mutable network_packet const *_payload;
    };

    /**
     * This class represents IPv6 packets.
     */
    class ipv6_packet : public ip_packet {
    public:
        /**
         * Constructor.
         *
         * @param data the data of the IPv6 packet.
         *
         * @param container the containing packet (if any).
         */
        ipv6_packet(packet_data const &data,
                    network_packet const *container = 0);

        /**
         * Virtual destructor.
         */
        virtual ~ipv6_packet();

        virtual void print_payload_info(ostream &out, size_t indent = 0) const;

        virtual ipv6_packet *shallow_copy() const;

        virtual bool good() const;
        virtual size_t header_size() const;
        virtual network_packet const *payload() const;

        virtual internet_address source_address(in_port_t port = 0) const;
        virtual internet_address dest_address(in_port_t port = 0) const;
    private:
        mutable network_packet const *_payload;
    };

    /**
     * This class represents UDP packets.  UDP packets are wrapped
     * inside IP (IPv4 or IPv6) packets.
     */
    class udp_packet : public network_packet {
    public:
        /**
         * Constructor.
         *
         * @param data the data of the UDP packet.
         *
         * @param container the containing IP packet.
         */
        udp_packet(packet_data const &data, ip_packet const *container);

        /**
         * Virtual destructor.
         */
        virtual ~udp_packet();

        virtual void print_payload_info(ostream &out, size_t indent = 0) const;

        virtual udp_packet *shallow_copy() const;

        virtual bool good() const;
        virtual size_t header_size() const;
        virtual network_packet const *payload() const;

        /**
         * The source address (including port number).
         */
        internet_address source_address() const;

        /**
         * The destination address (including port number).
         */
        internet_address dest_address() const;

        /**
         * The source port number.
         */
        in_port_t source_port() const;

        /**
         * The destination port number.
         */
        in_port_t dest_port() const;

        /**
         * Access the source address field as a value_ref.
         */
        value_ref field_src() const;

        /**
         * Access the destination address field as a value_ref.
         */
        value_ref field_dst() const;

        /**
         * Access the source port field as a value_ref.
         */
        value_ref field_src_port() const;

        /**
         * Access the destination port field as a value_ref.
         */
        value_ref field_dst_port() const;

    private:
        mutable network_packet const *_payload;
    };
}

#ifdef HAVE_EXT_HASH_MAP
namespace __gnu_cxx {
    using dns_analyzer::internet_address;

    /**
     * The hash of a dns_analyzer::internet_address.
     */
    template<> struct hash<internet_address> {

        /**
         * Calculate the hash.
         *
         * @param value the internet_address to hash.
         */
        size_t operator()(internet_address const &value) const {
            return value.hash_func();
        }
    };
}
#endif // HAVE_EXT_HASH_MAP

#endif // PACKET_HH
