// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef POINTERS_HH
#define POINTERS_HH

#include <algorithm>

namespace dns_analyzer {

    using std::swap;

    template <typename T>
    class scoped_ptr {
    public:
        explicit scoped_ptr(T *value = 0) : _value(value) {}
        ~scoped_ptr() {
            delete _value;
        }

        void reset(T *value = 0) {
            if (_value != value) {
                delete _value;
                _value = value;
            }
        }

        bool operator==(scoped_ptr<T> const &rhs) const {
            return _value == rhs._value;
        }

        bool operator!=(scoped_ptr<T> const &rhs) const {
            return _value != rhs._value;
        }

        operator bool() const { return _value != 0; }
        bool operator! () const { return _value == 0; }

        operator T *() const { return _value; };

        T &operator* () const { return *_value; }
        T *operator->() const { return _value; }

        void swap(scoped_ptr<T> &other) {
            swap(_value, other._value);
        }

    private:
        T *_value;

        // Do not allow copying and assignment.
        scoped_ptr(scoped_ptr<T> const &rhs);
        scoped_ptr<T> const &operator=(scoped_ptr<T> const &rhs);
    };

    template <typename T>
    class counted_ptr {
    public:
        explicit counted_ptr(T const *value = 0) : _count(new size_t(1)), _value(value) {}
        counted_ptr(counted_ptr<T> const &rhs) : _count(rhs._count), _value(rhs._value) {
            ++(*_count);
        }

        ~counted_ptr() {
            if (--(*_count) == 0) {
                delete _count;
                delete _value;
            }
        }

        counted_ptr<T> const &operator=(counted_ptr<T> const &rhs) {
            if (this != &rhs && _value != rhs._value) {
                if (--(*_count) == 0) {
                    delete _count;
                    delete _value;
                }
                _count = rhs._count;
                _value = rhs._value;
                ++(*_count);
            }

            return *this;
        }

        bool operator==(counted_ptr<T> const &rhs) const {
            return _value == rhs._value;
        }

        bool operator!=(counted_ptr<T> const &rhs) const {
            return _value != rhs._value;
        }

        operator bool() const { return _value != 0; }
        bool operator! () const { return _value == 0; }

        operator T *() const { return _value; };

        T &operator* () const { return *_value; }
        T *operator->() const { return _value; }

        void swap(scoped_ptr<T> &other) {
            swap(_count, other._count);
            swap(_value, other._value);
        }
    private:
        size_t *_count;
        T *_value;
    };
}

#endif // POINTERS_HH
