// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "r-project.hh"

#include "packet-dns.hh"
#include "util.hh"

#include <map>
#include <sstream>
#include <stdexcept>

namespace dns_analyzer {

    using std::dec;
    using std::hex;
    using std::fixed;
    using std::ios;
    using std::logic_error;
    using std::map;
    using std::ostringstream;
    using std::swap;

    template<typename T> static size_t
    unique_id(T const &val) {
        static size_t id = 0;
        static map<T, size_t> table;

        typename map<T, size_t>::const_iterator elem = table.find(val);
        if (elem == table.end()) {
            ++id;
            table[val] = id;
            return id;
        } else {
            return (*elem).second;
        }
    }

    r_project_listener_type::r_project_listener_type()
        : _base_filename(),
          _split_count(0),
          _output_count(0),
          _output()
    {
    }

    r_project_listener_type::~r_project_listener_type() {
        _output.close();
    }

    void
    r_project_listener_type::initialize(analyzer_type &/*analyzer*/) {
        _output_count = 0;
    }

    void
    r_project_listener_type::finalize(analyzer_type &/*analyzer*/) {
        _output_count = 0;
    }

    void
    r_project_listener_type::analyze(analyzer_type &analyzer,
                                     timeval const &timestamp,
                                     network_packet const &packet)
    {
        open_output();
        write_output(analyzer, timestamp, packet);
    }

    void
    r_project_listener_type::open_output() {
        if ((_split_count == 0 && _output_count == 0)
            || (_split_count != 0
                && (_output_count % _split_count) == 0))
        {
            if (_output.is_open()) {
                _output.close();
            }
            ostringstream filename;
            filename << _base_filename;
            if (_split_count > 0) {
                filename << '.';
                filename.width(3);
                filename.fill('0');
                filename << dec << (_output_count / _split_count);
            }
            filename << ".data";

            _output.open(filename.str().c_str());
            if (!_output) {
                throw logic_error("r_project_listener_type::open_output: open failed");
            }
            _output
                << "query timestamp "
                << "ip.client ip.server udp.client.port udp.server.port "
                << "dns.id dns.opcode dns.rcode dns.qcount dns.class dns.type "
                << "dns.domain\n";
        }
    }

    void
    r_project_listener_type::write_output(analyzer_type const &analyzer,
                                          timeval const &timestamp,
                                          network_packet const &packet)
    {
        try {
            udp_packet const *udp = find_payload<udp_packet>(&packet);
            dns_packet const *dns = find_payload<dns_packet>(&packet);

            if (dns == 0) {
                return;
            }

            resource_record_type rr(*dns->begin());

            _output << (dns->is_query() ? "T " : "F ")
                    << fixed << timeval_to_double(timestamp) << ' ';

            internet_address client_address = udp->source_address();
            internet_address server_address = udp->dest_address();

            if (!dns->is_query()) {
                swap(client_address, server_address);
            }

            if (analyzer.anonymize()) {
                _output << unique_id(client_address.to_string()) << ' '
                        << unique_id(server_address.to_string()) << ' ';
            } else {
                write_quoted_string(_output, client_address.to_string());
                _output << ' ';
                write_quoted_string(_output, server_address.to_string());
                _output << ' ';
            }
            _output << client_address.port() << ' '
                    << server_address.port() << ' '
                    << dns->id() << ' '
                    << dns->opcode() << ' '
                    << dns->rcode() << ' '
                    << dns->question_count() << ' '
                    << rr.the_class() << ' '
                    << rr.type() << ' ';
            if (analyzer.anonymize()) {
                _output << unique_id(rr.name().to_string());
            } else {
                write_quoted_string(_output, rr.name().to_string());
            }
        } catch (...) {
            _output << '\n';
            throw;
        }
        _output << '\n';
        ++_output_count;
    }
}
