// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "timer.hh"

#include "util.hh"

#include <iomanip>

namespace dns_analyzer {

    void
    timer_type::start() {
        gettimeofday(&_start_time, 0);
        getrusage(RUSAGE_SELF, &_start_rusage);
    }

    void
    timer_type::stop() {
        gettimeofday(&_stop_time, 0);
        getrusage(RUSAGE_SELF, &_stop_rusage);
    }

    static void
    elapsed(timeval &result,
            timeval const &start,
            timeval const &stop)
    {
        result.tv_sec = stop.tv_sec - start.tv_sec;
        result.tv_usec = stop.tv_usec - start.tv_usec;
        if (result.tv_usec < 0) {
            --result.tv_sec;
            result.tv_usec += 1000000;
        }
    }

    void
    timer_type::real_time(timeval &result) const {
        elapsed(result, _start_time, _stop_time);
    }

    double
    timer_type::real_time() const {
        timeval tv;
        real_time(tv);
        return timeval_to_double(tv);
    }

    void
    timer_type::cpu_time(timeval &result) const {
        timeval user;
        user_time(user);
        timeval sys;
        sys_time(sys);
        result.tv_sec = user.tv_sec + sys.tv_sec;
        result.tv_usec = user.tv_usec + sys.tv_usec;
        if (result.tv_usec > 1000000) {
            ++result.tv_sec;
            result.tv_usec -= 1000000;
        }
    }

    double
    timer_type::cpu_time() const {
        timeval tv;
        cpu_time(tv);
        return timeval_to_double(tv);
    }

    void
    timer_type::user_time(timeval &result) const {
        elapsed(result, _start_rusage.ru_utime, _stop_rusage.ru_utime);
    }

    double
    timer_type::user_time() const {
        timeval tv;
        user_time(tv);
        return timeval_to_double(tv);
    }

    void
    timer_type::sys_time(timeval &result) const {
        elapsed(result, _start_rusage.ru_stime, _stop_rusage.ru_stime);
    }

    double
    timer_type::sys_time() const {
        timeval tv;
        sys_time(tv);
        return timeval_to_double(tv);
    }

    void
    print_timeval(ostream &out, char const *label, timeval const &tv) {
        using namespace std;

        out.fill(' ');
        out.width(8);
        out.setf(ios_base::left, ios_base::adjustfield);
        out << label;
        out.fill('0');
        out.width(1);
        out << tv.tv_sec / 60 << 'm';
        out.width(1);
        out << tv.tv_sec % 60 << '.';
        out.width(3);
        out << tv.tv_usec / 1000 << 's' << endl;
    }

    void
    timer_type::print(ostream &out) {
        using namespace std;

        timeval tv;
        real_time(tv);
        print_timeval(out, "real", tv);
        cpu_time(tv);
        print_timeval(out, "cpu", tv);
        user_time(tv);
        print_timeval(out, "user", tv);
        sys_time(tv);
        print_timeval(out, "sys", tv);
    }
}
