// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef UTIL_HH
#define UTIL_HH

#include "fast_map.hh"

#include <iosfwd>
#include <string>
#include <vector>
#include <sys/time.h>

namespace dns_analyzer {
    using std::ostream;
    using std::string;

    /**
     * Convert a timeval structure to a double representing the time
     * in seconds (with microsecond precision)..
     */
    double timeval_to_double(timeval const &tv);

    /**
     * Double-quote the string and escape any non-printable characters
     * and double-quotes in the string.
     */
    string quote_string(string const &s);

    /**
     * Write a quoted string to the specified output stream.
     *
     * @param out the output stream.
     *
     * @param s the string to quote and write.
     *
     * @returns the output stream.
     */
    ostream &write_quoted_string(ostream &out, string const &s);
}

#ifdef HAVE_EXT_HASH_MAP
namespace __gnu_cxx {
    using std::allocator;
    using std::vector;

    /**
     * The hash of a bool.
     */
    template<> struct hash<bool> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(bool value) {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a double.
     */
    template<> struct hash<double> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(double value) {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a float.
     */
    template<> struct hash<float> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(float value) {
            return static_cast<size_t>(value);
        }
    };

    /**
     * The hash of a string.
     */
    template<> struct hash<std::string> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(std::string const &value) const {
            size_t result = 0;
            for (std::string::const_iterator it = value.begin();
                 it != value.end();
                 ++it)
            {
                result = 5 * result + static_cast<size_t>(*it);
            }
            return result;
        }
    };

    /**
     * The hash of a pointer.
     */
    template<typename T> struct hash<T *> {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(T *value) const {
            return reinterpret_cast<size_t>(value);
        }
    };

    /**
     * The hash of a vector of elements.
     */
    template<typename T> struct hash<vector<T, allocator<T> > > {

        /**
         * Calculate the hash value.
         *
         * @param value the value to hash.
         */
        size_t operator()(vector<T> const &value) const {
            hash<T> hasher;
            size_t result = 0;

            for (typename vector<T>::const_iterator it = value.begin();
                 it != value.end();
                 ++it)
            {
                result = 5 * result + hasher(*it);
            }
            return result;
        }
    };
}
#endif // HAVE_EXT_HASH_MAP

#endif // UTIL_HH
