// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "value.hh"

#include <cassert>
#include <typeinfo>

namespace dns_analyzer {

    using std::type_info;

    int
    value_type::compare(value_type const &rhs) const {
        if (this == &rhs) {
            return 0;
        } else if (rhs.is_null()) {
            return -rhs.compare(*this);
        }

        type_info const &lhs_type = typeid(*this);
        type_info const &rhs_type = typeid(rhs);

        if (lhs_type.before(rhs_type)) {
            return -1;
        } else if (lhs_type == rhs_type) {
            return 0;
        } else {
            return +1;
        }
    }

    inline void
    value_ref::add_ref() {
        ++_value->_ref_count;
    }

    inline void
    value_ref::remove_ref() {
        assert(_value->_ref_count > 0);
        --_value->_ref_count;
        if (_value->_ref_count == 0) {
            delete _value;
        }
    }

    value_ref::value_ref() : _value(&null_value) {
        add_ref();
    }

    value_ref::value_ref(value_type *value) : _value(value) {
        add_ref();
    }

    value_ref::value_ref(value_ref const &rhs) : _value(rhs._value) {
        add_ref();
    }

    value_ref::value_ref(value_type const &value) : _value(value.clone()) {
        add_ref();
    }

    value_ref::~value_ref() {
        remove_ref();
    }

    value_ref const &
    value_ref::operator=(value_ref const &rhs) {
        if (this != &rhs && _value != rhs._value) {
            remove_ref();
            _value = rhs._value;
            add_ref();
        }

        return *this;
    }

    value_ref const &
    value_ref::operator=(value_type const &rhs) {
        if (_value != &rhs) {
            _value = rhs.clone();
            add_ref();
        }

        return *this;
    }

    bool
    value_ref::operator==(value_ref const &rhs) {
        return *_value == *rhs._value;
    }

    bool
    value_ref::operator!=(value_ref const &rhs) {
        return *_value != *rhs._value;
    }

    bool
    value_ref::operator< (value_ref const &rhs) {
        return *_value < *rhs._value;
    }

    bool
    value_ref::operator<=(value_ref const &rhs) {
        return *_value <= *rhs._value;
    }

    bool
    value_ref::operator> (value_ref const &rhs) {
        return *_value > *rhs._value;
    }

    bool
    value_ref::operator>=(value_ref const &rhs) {
        return *_value >= *rhs._value;
    }

    int
    null_value_type::compare(value_type const &rhs) const {
        if (rhs.is_null()) {
            return 0;
        } else {
            return -1;
        }
    }

    null_value_type null_value_type::instance;

    null_value_type &null_value = null_value_type::instance;
}
