// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef VALUE_HH
#define VALUE_HH

#include "fast_map.hh"
#include "util.hh"

#include <ostream>
#include <sstream>
#include <string>
#include <vector>


namespace dns_analyzer {

    using std::ostream;
    using std::ostringstream;
    using std::string;
    using std::vector;

    class value_ref;

    /**
     * 
     */
    class value_type {
    public:
        value_type() : _ref_count(0) {}
        virtual ~value_type() {}

        virtual int compare(value_type const &rhs) const;

        virtual bool is_null() const { return false; }

        virtual value_type *clone() const = 0;
        virtual void print(ostream &out) const = 0;
        virtual size_t hash_func() const = 0;

        string to_string() const {
            ostringstream out;
            print(out);
            return out.str();
        }

    protected:
        explicit value_type(size_t ref_count) : _ref_count(ref_count) {}
        value_type const &operator=(value_type const &rhs);

    private:
        size_t _ref_count;

        friend class value_ref;
    };

    class null_value_type : public value_type {
    private:
        null_value_type() : value_type(1) {}
    public:
        virtual ~null_value_type() {}

        static null_value_type instance;

        null_value_type const &value();

        virtual int compare(value_type const &rhs) const;

        virtual bool is_null() const { return true; }

        virtual null_value_type *clone() const {
            return new null_value_type;
        }

        virtual void print(ostream &out) const {
            out << "null";
        }

        virtual size_t hash_func() const { return 0; }
    };

    class value_ref {
    public:
        value_ref();
        /**
         * @param rhs [function]
         */
        value_ref(value_ref const &rhs);

        /**
         * @param value [ownership transfer]
         */
        value_ref(value_type *value);

        value_ref(value_type const &value);

        ~value_ref();

        value_ref const &operator=(value_ref const &rhs);
        value_ref const &operator=(value_type const &rhs);

        operator value_type const &() const {
            return *_value;
        }

        value_type const &value() const {
            return *_value;
        }

        bool operator==(value_ref const &rhs);
        bool operator!=(value_ref const &rhs);
        bool operator< (value_ref const &rhs);
        bool operator<=(value_ref const &rhs);
        bool operator> (value_ref const &rhs);
        bool operator>=(value_ref const &rhs);

        size_t hash_func() const {
            return _value->hash_func();
        }
    private:
        value_type *_value;

        void add_ref();
        void remove_ref();
    };

    template <typename T>
    class typed_value : public value_type {
    public:
        explicit typed_value(T const &init) : _value(init) {}
        virtual ~typed_value() {}

        T const &value() const { return _value; }

        virtual int compare(value_type const &rhs) const {
            int result = value_type::compare(rhs);
            if (result != 0) {
                return result;
            }

            typed_value<T> const &typed_rhs
                = dynamic_cast<typed_value<T> const &>(rhs);
            if (_value < typed_rhs._value) {
                return -1;
            } else if (_value == typed_rhs._value) {
                return 0;
            } else {
                return +1;
            }
        }

        virtual typed_value<T> *clone() const {
            return new typed_value<T>(_value);
        }

        virtual void print(ostream &out) const {
            out << _value;
        }

        virtual size_t hash_func() const {
#ifdef HAVE_EXT_HASH_MAP
            __gnu_cxx::hash<T> hasher;
            return hasher(_value);
#else
            return 0;
#endif
        }
    private:
        T _value;
    };

    extern null_value_type &null_value;

    inline bool operator==(value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) == 0;
    }

    inline bool operator!=(value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) != 0;
    }

    inline bool operator< (value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) < 0;
    }

    inline bool operator<=(value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) <= 0;
    }

    inline bool operator> (value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) > 0;
    }

    inline bool operator>=(value_type const &lhs, value_type const &rhs) {
        return lhs.compare(rhs) >= 0;
    }

    inline ostream &operator<<(ostream &out, value_type const &rhs) {
        rhs.print(out);
        return out;
    }
}

#ifdef HAVE_EXT_HASH_MAP
namespace __gnu_cxx {
    using namespace dns_analyzer;

    template<> struct hash<value_type> {
        size_t operator()(value_type const &value) const { return value.hash_func(); }
    };
    template<> struct hash<value_ref> {
        size_t operator()(value_ref const &ref) const { return ref.hash_func(); }
    };
}
#endif // HAVE_EXT_HASH_MAP

#endif // VALUE_HH
