// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <config.h>

#include "variable.hh"

#include "analyzer.hh"
#include "util.hh"
#include "value.hh"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <limits>
#include <ostream>
#include <stdexcept>
#include <iostream>

namespace dns_analyzer {

    using std::pair;
    using std::cerr;
    using std::dec;
    using std::max;
    using std::sort;

    stat_type::stat_type()
        : _count(0), _sum(0.0), _sum_of_squares(0.0), _modified_at(0)
    {
    }

    void
    stat_type::add(float const &value) {
        ++_count;
        _sum += value;
        _sum_of_squares += value * value;
    }

    float
    stat_type::variance() const {
        return (sum_of_squares() - sum() * sum() / count()) / (count() - 1);
    }

    float
    stat_type::stdev() const {
        return sqrt(variance());
    }

    ostream &
    operator<<(ostream &out, stat_type const &rhs) {
        return out << "count: " << rhs.count()
                   << ", sum: " << rhs.sum()
                   << ", avg: " << rhs.average()
                   << ", sd: " << rhs.stdev();
    }

    interval_stat_type::interval_stat_type(stat_type &stat,
                                           float time_interval)
        : _stat(&stat),
          _time_interval(time_interval),
          _last_update_timestamp(current_packet_timestamp),
          _last_count(0),
          _last_sum(0.0),
          _last_sum_of_squares(0.0)
    {
    }

    void
    interval_stat_type::update(float timestamp) {
        float interval = timestamp - _last_update_timestamp;
        if (interval >= _time_interval) {
            _last_update_timestamp = timestamp;

            _last_count = _stat->count();
            _last_sum = _stat->sum();
            _last_sum_of_squares = _stat->sum_of_squares();
        }
    }

    combined_stat_config_type::combined_stat_config_type(
        vector<float> const &intervals)
        : _intervals(intervals)
    {
        assert(!_intervals.empty());
        sort(_intervals.begin(), _intervals.end());
    }

    combined_stat_type::combined_stat_type()
        : _stat(), _terms(0)
    {
    }

    combined_stat_type::combined_stat_type(combined_stat_type const &rhs)
        : _stat(rhs._stat), _terms(0)
    {
        if (rhs._terms != 0) {
            _terms.reset(new vector<interval_stat_type>(*rhs._terms));

            for (vector<interval_stat_type>::iterator it = _terms->begin();
                 it != _terms->end();
                 ++it)
            {
                (*it).stat(_stat);
            }
        }
    }

    combined_stat_type const &
    combined_stat_type::operator=(combined_stat_type const &rhs) {
        if (this != &rhs) {
            _stat = rhs._stat;
            if (rhs._terms == 0) {
                _terms.reset(0);
            } else {
                _terms.reset(new vector<interval_stat_type>(*rhs._terms));
                for (vector<interval_stat_type>::iterator it = _terms->begin();
                     it != _terms->end();
                     ++it)
                {
                    (*it).stat(_stat);
                }
            }
        }

        return *this;
    }

    void
    combined_stat_type::configure(combined_stat_config_type const *config)
    {
        if (config == 0 || _terms != 0) {
            return;
        }

        _terms.reset(new vector<interval_stat_type>());
        _terms->reserve(config->intervals().size());
        for (vector<float>::const_iterator it = config->intervals().begin();
             it != config->intervals().end();
             ++it)
        {
            _terms->push_back(interval_stat_type(_stat, *it));
        }
    }

    void
    combined_stat_type::update(float timestamp) {
        if (_terms != 0) {
            for (vector<interval_stat_type>::iterator it = _terms->begin();
                 it != _terms->end();
                 ++it)
            {
                (*it).update(timestamp);
            }
        }
    }

    variable_type::variable_type()
        : _add_count(0), _group_by(), _stats()
    {
    }

    variable_type::variable_type(group_names const &groups)
        : _add_count(0), _group_by(groups), _stats()
    {
    }

    variable_type::~variable_type() {
    }

    variable_type::group_stat_type::mapped_type &
    variable_type::add(group_values const &group, float value) {
        assert(group.size() == _group_by.size());
        ++_add_count;
        group_stat_type::mapped_type &mapped = _stats[group];
        mapped.add(value);
        mapped.modified_at(_add_count);
        if (_stat_config != 0) {
            mapped.configure(_stat_config);
            _updated[&mapped] = group;
        }
        return mapped;
    }

    void
    variable_type::group_by(group_names const &value) {
        _group_by = value;
    }

    stat_type
    variable_type::global_stats() const {
        stat_type result;

        for (group_stat_type::const_iterator it = _stats.begin();
             it != _stats.end();
             ++it)
        {
            result.add((*it).second.sum());
        }

        return result;
    }

    variable_type::group_stat_type::mapped_type const &
    variable_type::stats(group_values const &group) {
        return _stats[group];
    }

    void
    variable_type::prune(size_t threshold) {
        size_t count(0);
        size_t total_count(_stats.size());

        if (total_count == 0) {
            return;
        }

        group_stat_type::iterator it = _stats.begin();
        while (it != _stats.end()) {
            group_stat_type::mapped_type &s = (*it).second;
            size_t age = _add_count - s.modified_at();
            if (age / s.count() > threshold) {
                group_stat_type::iterator current = it;
                ++it;
                ++count;
                _updated.erase(&s);
                _stats.erase(current);
            } else {
                ++it;
            }
        }

//         cerr << "pruning: " << dec
//              << count << " of " << total_count << " ("
//              << 100 * count / total_count
//              << "%) pruned\n";
    }

    static void
    print_interval_stat(ostream &out, float timestamp,
                        interval_stat_type const &rhs)
    {
        float interval = timestamp - rhs.last_update_timestamp();
        out << '\t' << dec << rhs.interval_sum() << " in "
            << interval << " seconds ("
            << (rhs.interval_sum() / interval)
            << "/s)\n";
    }

    void
    variable_type::configure(combined_stat_config_type const *stat_config) {
        _stat_config = stat_config;
    }

    void
    variable_type::update(float timestamp) {
        if (_stat_config == 0 || _updated.size() == 0) {
            return;
        }

        cerr << "Updated: " << _updated.size() << '\n';

        updated_table_type::iterator largest = _updated.begin();
        for (updated_table_type::iterator it = _updated.begin();
             it != _updated.end();
             ++it)
        {
            interval_stat_type const &a = (*largest).first->terms()->at(0);
            interval_stat_type const &b = (*it).first->terms()->at(0);
            float a_dt = timestamp - a.last_update_timestamp();
            float b_dt = timestamp - b.last_update_timestamp();
            if ((*it).first->count() >= 10
                && (a.interval_sum() / a_dt < b.interval_sum() / b_dt))
            {
                largest = it;
            }
        }

        if (largest != _updated.end()) {
            cerr << timestamp << ": ";
            print_group_values(cerr, (*largest).second);
            cerr << '\n';
            group_stat_type::mapped_type const &s = *(*largest).first;
            for (vector<interval_stat_type>::const_iterator it = s.terms()->begin();
                 it != s.terms()->end();
                 ++it)
            {
                print_interval_stat(cerr, timestamp, *it);
            }
        }

        for (updated_table_type::iterator it = _updated.begin();
             it != _updated.end();
             ++it)
        {
            group_stat_type::mapped_type *s = (*it).first;
            s->update(timestamp);
        }

        _updated.clear();
    }

    void print_group_values(ostream &out,
                            variable_type::group_values const &values)
    {
        for (size_t j = 0; j < values.size(); ++j) {
            ostringstream temp;
            temp << values.at(j);
            write_quoted_string(out, temp.str());
            if (j < values.size() - 1) {
                out << ' ';
            }
        }
    }
}
