// Copyright (C) 2002, 2003, NLnet Labs, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef VARIABLE_HH
#define VARIABLE_HH

#include "fast_map.hh"
#include "pointers.hh"
#include "util.hh"

#include <iosfwd>
#include <limits>
#include <map>
#include <string>
#include <vector>

namespace dns_analyzer {

    using std::map;
    using std::numeric_limits;
    using std::ostream;
    using std::string;
    using std::vector;

    class value_type;
    class value_ref;

    /**
     * This class calculates some statistical information over a
     * number series.
     */
    class stat_type {
    public:
        /**
         * Constructor.
         */
        stat_type();

        /**
         * Add the value to the number series.
         */
        void add(float const &value);

        /**
         * The count of numbers in the number series.
         */
        unsigned count() const { return _count; }

        /**
         * The sum of the number series.
         */
        float sum() const { return _sum; }

        /**
         * The sum of squares of the number series.
         */
        float sum_of_squares() const { return _sum_of_squares; }

        /**
         * The average of the number series.
         */
        float average() const { return _sum / _count; }

        /**
         * The variance of the number series.
         */
        float variance() const;

        /**
         * The standard deviation of the number series.
         */
        float stdev() const;

        /**
         * Set the last time this stat was modified.
         *
         * @param value the time this stat was modified.  Value should
         * be a strictly increasing number (but does not need any
         * relation to "real" time).
         */
        void modified_at(size_t value) { _modified_at = value; }

        /**
         * The time this stat was last modified.  See modified_at(size_t).
         */
        size_t modified_at() const { return _modified_at; }

    private:
        unsigned _count;
        float _sum;
        float _sum_of_squares;
        size_t _modified_at;
    };

    ostream &operator<<(ostream &out, stat_type const &rhs);

    /**
     * Measure data over a time interval.
     */
    class interval_stat_type {
    public:
        explicit interval_stat_type(
            stat_type &stat,    // lifetime: instance
            float time_interval = numeric_limits<float>::infinity());

        /**
         * Update interval statistics if the provided timestamp >=
         * last_update_timestamp + time_interval.
         */
        void update(float timestamp);

        unsigned interval_count() const {
            return _stat->count() - _last_count;
        }
        float interval_sum() const {
            return _stat->sum() - _last_sum;
        }
        float interval_sum_of_squares() const {
            return _stat->sum_of_squares() - _last_sum_of_squares;
        }

        float time_interval() const { return _time_interval; }
        float last_update_timestamp() const { return _last_update_timestamp; }

        void stat(stat_type &value) { _stat = &value; }

        stat_type &stat() { return *_stat; }
        stat_type const &stat() const { return *_stat; }
    private:
        stat_type *_stat;

        float _time_interval;
        float _last_update_timestamp;

        unsigned _last_count;
        float _last_sum;
        float _last_sum_of_squares;
    };

    /**
     * Write a string representation of an interval_stat_type to the
     * specified ostream.
     *
     * @param out the output stream.
     *
     * @param rhs the interval_stat_type instance.
     *
     * @returns the output stream.
     */
    ostream &operator<<(ostream &out, interval_stat_type const &rhs);

    /**
     * This class represents the configuration of a
     * combined_stat_type.  Currently only the time intervals for
     * throughput measurement can be configured.
     */
    class combined_stat_config_type {
    public:
        /**
         * Construct a new instance with the provided time intervals.
         *
         * @param intervals the time intervals.  This vector must not
         * be empty.
         */
        explicit combined_stat_config_type(vector<float> const &intervals);

        /**
         * The configured time intervals.
         */
        vector<float> const &intervals() const { return _intervals; }

    private:
        vector<float> _intervals;
    };

    /**
     * Measure statistics and the modification of these statistics
     * over time.  This class combines a stat_type instances with an
     * optional vector of interval_stat_type instances.
     */
    class combined_stat_type {
    public:
        /**
         * Default constructor.
         */
        combined_stat_type();

        /**
         * Copy constructor.
         */
        combined_stat_type(combined_stat_type const &rhs);

        /**
         * Assignment.
         */
        combined_stat_type const &operator=(combined_stat_type const &rhs);

        /**
         * Configure this instance.  An instance can only be
         * configured once.  If the configuration information is not
         * null the interval_stat_type instances will be initialized
         * according to the time intervals stored in the
         * configuration.
         *
         * @param config the configuration information [function].
         */
        void configure(combined_stat_config_type const *config);

        void add(float value) { _stat.add(value); }
        unsigned count() const { return _stat.count(); }
        float sum() const { return _stat.sum(); }

        void modified_at(size_t value) { _stat.modified_at(value); }
        size_t modified_at() const { return _stat.modified_at(); }

        void update(float timestamp);

        stat_type &stat() { return _stat; }
        stat_type const &stat() const { return _stat; }

        vector<interval_stat_type> const *terms() const { return _terms; }
    private:
        stat_type _stat;
        scoped_ptr<vector<interval_stat_type> > _terms;
    };

    /**
     * A "statistical" variable.  The variable gathers information for
     * each item added such as the number of items, the sum of the
     * items, the squared sum of the items, etc.
     */
    class variable_type {
    public:
        typedef vector<string> group_names;
        typedef vector<value_ref> group_values;
        typedef fast_map<group_values, combined_stat_type> group_stat_type;
        typedef fast_map<combined_stat_type *, group_values> updated_table_type;

        variable_type();
        explicit variable_type(group_names const &group_by);
        ~variable_type();

        group_stat_type::mapped_type &add(group_values const &group, float value);

        void group_by(group_names const &value);
        group_names const &group_by() const { return _group_by; }

        stat_type global_stats() const;
        group_stat_type const &stats() const {
            return _stats;
        }
        group_stat_type::mapped_type const &stats(group_values const &group);

        /**
         * Prune "uninteresting" data items.  Currently all items that
         * have age divided by (4+count()) greater than
         * 'threshold' are removed.
         */
        void prune(size_t threshold);

        void configure(combined_stat_config_type const *stat_config);
        void update(float timestamp);
    private:
        size_t _add_count;
        group_names _group_by;
        group_stat_type _stats;
        combined_stat_config_type const *_stat_config;
        updated_table_type _updated;

        // Disallow copying.
        variable_type(variable_type const &rhs);
        variable_type &operator=(variable_type const &rhs);
    };

    void print_group_values(ostream &out,
                            variable_type::group_values const &values);
}

#endif // VARIABLE_HH
